import shutil
from pathlib import Path
from typing import Any

import pytest
import typer

from desktop_2fa.cli import helpers
from desktop_2fa.vault.vault import UnsupportedFormat, VaultIOError

TEST_PASSWORD = "jawislajawisla"


@pytest.fixture
def fake_vault_env_helpers(tmp_path: Path, monkeypatch: Any) -> Path:
    fake_vault = tmp_path / "vault"

    monkeypatch.setattr(
        "desktop_2fa.cli.helpers.get_vault_path",
        lambda: str(fake_vault),
    )

    if fake_vault.parent.exists():
        shutil.rmtree(fake_vault.parent)
    fake_vault.parent.mkdir(parents=True, exist_ok=True)

    return fake_vault


def test_helpers_add_and_list_entries(
    fake_vault_env_helpers: Path, capsys: Any
) -> None:
    from desktop_2fa.vault import Vault

    helpers.save_vault(fake_vault_env_helpers, Vault(), TEST_PASSWORD)
    helpers.add_entry(
        fake_vault_env_helpers,
        "GitHub",
        "GitHub",
        "JBSWY3DPEHPK3PXP",
        TEST_PASSWORD,
    )

    out = capsys.readouterr().out.strip().splitlines()
    # Najpierw "Added entry: GitHub"
    assert out == ["Added entry: GitHub"]

    helpers.list_entries(fake_vault_env_helpers, TEST_PASSWORD)
    out = capsys.readouterr().out.strip().splitlines()
    assert out == ["- GitHub (GitHub)"]

    vault = helpers.load_vault(fake_vault_env_helpers, TEST_PASSWORD)
    assert len(vault.entries) == 1
    assert vault.entries[0].issuer == "GitHub"
    assert vault.entries[0].account_name == "GitHub"


def test_helpers_generate_code(fake_vault_env_helpers: Path, capsys: Any) -> None:
    from desktop_2fa.vault import Vault

    helpers.save_vault(fake_vault_env_helpers, Vault(), TEST_PASSWORD)
    helpers.add_entry(
        fake_vault_env_helpers,
        "GitHub",
        "GitHub",
        "JBSWY3DPEHPK3PXP",
        TEST_PASSWORD,
    )
    capsys.readouterr()  # wyczyść output po add_entry

    helpers.generate_code(fake_vault_env_helpers, "GitHub", TEST_PASSWORD)
    out = capsys.readouterr().out.strip()

    lines = out.splitlines()
    code = lines[-1]
    assert len(code) == 6
    assert code.isdigit()


def test_helpers_remove_entry(fake_vault_env_helpers: Path) -> None:
    from desktop_2fa.vault import Vault

    helpers.save_vault(fake_vault_env_helpers, Vault(), TEST_PASSWORD)
    helpers.add_entry(
        fake_vault_env_helpers,
        "GitHub",
        "GitHub",
        "JBSWY3DPEHPK3PXP",
        TEST_PASSWORD,
    )
    helpers.remove_entry(fake_vault_env_helpers, "GitHub", TEST_PASSWORD)

    vault = helpers.load_vault(fake_vault_env_helpers, TEST_PASSWORD)
    assert len(vault.entries) == 0


def test_helpers_rename_entry(fake_vault_env_helpers: Path) -> None:
    from desktop_2fa.vault import Vault

    helpers.save_vault(fake_vault_env_helpers, Vault(), TEST_PASSWORD)
    helpers.add_entry(
        fake_vault_env_helpers,
        "GitHub",
        "GitHub",
        "JBSWY3DPEHPK3PXP",
        TEST_PASSWORD,
    )
    helpers.rename_entry(fake_vault_env_helpers, "GitHub", "GitHub2", TEST_PASSWORD)

    vault = helpers.load_vault(fake_vault_env_helpers, TEST_PASSWORD)
    assert vault.entries[0].issuer == "GitHub2"
    assert vault.entries[0].account_name == "GitHub2"


def test_helpers_export_and_import(
    fake_vault_env_helpers: Path, tmp_path: Path
) -> None:
    from desktop_2fa.vault import Vault

    helpers.save_vault(fake_vault_env_helpers, Vault(), TEST_PASSWORD)
    helpers.add_entry(
        fake_vault_env_helpers,
        "GitHub",
        "GitHub",
        "JBSWY3DPEHPK3PXP",
        TEST_PASSWORD,
    )

    tmp_file = tmp_path / "export.bin"
    helpers.export_vault(fake_vault_env_helpers, tmp_file, TEST_PASSWORD)

    vault = helpers.load_vault(fake_vault_env_helpers, TEST_PASSWORD)
    vault.entries.clear()
    helpers.save_vault(fake_vault_env_helpers, vault, TEST_PASSWORD)

    helpers.import_vault(fake_vault_env_helpers, tmp_file, TEST_PASSWORD)
    vault = helpers.load_vault(fake_vault_env_helpers, TEST_PASSWORD)

    assert len(vault.entries) == 1
    assert vault.entries[0].issuer == "GitHub"


def test_helpers_backup(fake_vault_env_helpers: Path) -> None:
    from desktop_2fa.vault import Vault

    helpers.save_vault(fake_vault_env_helpers, Vault(), TEST_PASSWORD)
    helpers.add_entry(
        fake_vault_env_helpers,
        "GitHub",
        "GitHub",
        "JBSWY3DPEHPK3PXP",
        TEST_PASSWORD,
    )

    backup_path = fake_vault_env_helpers.with_suffix(".backup.bin")
    helpers.backup_vault(fake_vault_env_helpers, backup_path, TEST_PASSWORD)

    assert backup_path.exists()
    assert backup_path.stat().st_size > 0


def test_helpers_export_vault_missing(
    fake_vault_env_helpers: Path, tmp_path: Path, capsys: Any
) -> None:
    from desktop_2fa.vault import Vault

    helpers.save_vault(fake_vault_env_helpers, Vault(), TEST_PASSWORD)
    export_path = tmp_path / "export.bin"
    helpers.export_vault(fake_vault_env_helpers, export_path, TEST_PASSWORD)
    out = capsys.readouterr().out
    # Aktualne zachowanie: eksportuje istniejący vault
    assert "Exported vault to:" in out
    assert export_path.exists()


def test_helpers_import_vault_missing(
    fake_vault_env_helpers: Path, tmp_path: Path
) -> None:
    from desktop_2fa.vault import Vault

    helpers.save_vault(fake_vault_env_helpers, Vault(), TEST_PASSWORD)
    missing = tmp_path / "nope.bin"
    with pytest.raises(VaultIOError):
        helpers.import_vault(fake_vault_env_helpers, missing, TEST_PASSWORD)


def test_helpers_backup_vault_missing(
    fake_vault_env_helpers: Path, capsys: Any
) -> None:
    from desktop_2fa.vault import Vault

    helpers.save_vault(fake_vault_env_helpers, Vault(), TEST_PASSWORD)

    backup_path = fake_vault_env_helpers.with_suffix(".backup.bin")
    helpers.backup_vault(fake_vault_env_helpers, backup_path, TEST_PASSWORD)
    out = capsys.readouterr().out
    # Aktualne zachowanie: backup_vault pisze backup i drukuje "Backup created:"
    assert "Backup created:" in out
    assert backup_path.exists()


def test_get_password_for_vault_password_file_missing(
    tmp_path: Path, capsys: Any
) -> None:
    import typer

    fake_ctx = type(
        "FakeContext",
        (),
        {"obj": {"password_file": str(tmp_path / "missing.txt"), "interactive": True}},
    )()
    with pytest.raises(typer.Exit):
        helpers.get_password_for_vault(fake_ctx, new_vault=False)
    out = capsys.readouterr().out
    assert "Error: Password file" in out


def test_load_vault_failed(tmp_path: Path) -> None:
    fake_vault = tmp_path / "vault"
    fake_vault.write_text("invalid")
    with pytest.raises(UnsupportedFormat):
        helpers.load_vault(fake_vault, TEST_PASSWORD)


def test_get_vault_path() -> None:
    path = helpers.get_vault_path()
    assert isinstance(path, str)
    assert ".desktop-2fa" in path


def test_get_password_for_vault_passwords_not_match(monkeypatch: Any) -> None:
    import typer

    fake_ctx = type("FakeContext", (), {"obj": {"interactive": True}})()
    # Mock to return different passwords for the two prompts
    responses = ["pass1", "pass2"]
    monkeypatch.setattr("typer.prompt", lambda text, hide_input: responses.pop(0))
    with pytest.raises(typer.Exit):
        helpers.get_password_for_vault(fake_ctx, new_vault=True)


def test_get_password_for_vault_both_password_and_file(
    tmp_path: Path, capsys: Any
) -> None:
    """Test error when both password and password_file are provided."""
    import typer

    fake_ctx = type(
        "FakeContext",
        (),
        {
            "obj": {
                "password": "test",
                "password_file": str(tmp_path / "test.txt"),
                "interactive": True,
            }
        },
    )()

    with pytest.raises(typer.Exit):
        helpers.get_password_for_vault(fake_ctx, new_vault=False)

    out = capsys.readouterr().out
    assert "Error: Cannot specify both --password and --password-file" in out


def test_get_password_for_vault_no_password_non_interactive(capsys: Any) -> None:
    """Test error when no password provided and not in interactive mode."""
    import typer

    fake_ctx = type("FakeContext", (), {"obj": {"interactive": False}})()

    with pytest.raises(typer.Exit):
        helpers.get_password_for_vault(fake_ctx, new_vault=False)

    out = capsys.readouterr().out
    assert "Error: Password not provided and not running in interactive mode" in out


def test_print_entries_table_empty(capsys: Any) -> None:
    """Test print_entries_table with empty list."""
    helpers.print_entries_table([])

    out = capsys.readouterr().out
    assert "No entries found." in out


def test_validate_base32_valid() -> None:
    """Test validate_base32 with valid Base32 strings."""
    assert helpers.validate_base32("JBSWY3DPEHPK3PXP") is True
    assert helpers.validate_base32("JBSWY3DP EHPK3PXP") is True  # with spaces
    assert helpers.validate_base32("jbswy3dpehpk3pxp") is True  # lowercase
    assert helpers.validate_base32("JBSWY3DPEHPK3PXP=") is True  # with padding


def test_validate_base32_invalid() -> None:
    """Test validate_base32 with invalid Base32 strings."""
    assert helpers.validate_base32("invalid") is True  # Actually valid Base32
    assert helpers.validate_base32("JBSWY3DPEHPK3PXP1") is False  # invalid character
    assert helpers.validate_base32("JBSWY3DPEHPK3PXP!") is False  # special character


def test_parse_otpauth_url_valid() -> None:
    """Test parse_otpauth_url with valid URLs."""
    result = helpers.parse_otpauth_url(
        "otpauth://totp/GitHub:octocat?secret=JBSWY3DPEHPK3PXP&issuer=GitHub"
    )
    assert result["issuer"] == "GitHub"
    assert result["label"] == "octocat"
    assert result["secret"] == "JBSWY3DPEHPK3PXP"


def test_parse_otpauth_url_issuer_only() -> None:
    """Test parse_otpauth_url with issuer only."""
    result = helpers.parse_otpauth_url("otpauth://totp/GitHub?secret=JBSWY3DPEHPK3PXP")
    assert result["issuer"] == "GitHub"
    assert result["label"] == "GitHub"
    assert result["secret"] == "JBSWY3DPEHPK3PXP"


def test_parse_otpauth_url_label_only() -> None:
    """Test parse_otpauth_url with label only."""
    result = helpers.parse_otpauth_url(
        "otpauth://totp/:octocat?secret=JBSWY3DPEHPK3PXP"
    )
    assert result["issuer"] == "Unknown"
    assert result["label"] == "octocat"
    assert result["secret"] == "JBSWY3DPEHPK3PXP"


def test_parse_otpauth_url_no_colon() -> None:
    """Test parse_otpauth_url with no colon in path."""
    result = helpers.parse_otpauth_url("otpauth://totp/GitHub?secret=JBSWY3DPEHPK3PXP")
    assert result["issuer"] == "GitHub"
    assert result["label"] == "GitHub"
    assert result["secret"] == "JBSWY3DPEHPK3PXP"


def test_parse_otpauth_url_invalid_scheme() -> None:
    """Test parse_otpauth_url with invalid scheme."""
    with pytest.raises(ValueError, match="Invalid otpauth URL"):
        helpers.parse_otpauth_url("http://example.com")


def test_parse_otpauth_url_invalid_type() -> None:
    """Test parse_otpauth_url with non-TOTP type."""
    with pytest.raises(ValueError, match="Only TOTP otpauth URLs are supported"):
        helpers.parse_otpauth_url("otpauth://hotp/GitHub?secret=JBSWY3DPEHPK3PXP")


def test_parse_otpauth_url_missing_secret() -> None:
    """Test parse_otpauth_url with missing secret."""
    with pytest.raises(ValueError, match="Secret parameter is required"):
        helpers.parse_otpauth_url("otpauth://totp/GitHub?issuer=GitHub")


def test_parse_otpauth_url_issuer_in_query() -> None:
    """Test parse_otpauth_url with issuer in query parameter."""
    result = helpers.parse_otpauth_url(
        "otpauth://totp/:octocat?secret=JBSWY3DPEHPK3PXP&issuer=GitHub"
    )
    assert result["issuer"] == "GitHub"
    assert result["label"] == "octocat"
    assert result["secret"] == "JBSWY3DPEHPK3PXP"


def test_print_success(capsys: Any) -> None:
    """Test print_success function."""
    helpers.print_success("Test message")
    out = capsys.readouterr().out
    assert "Test message" in out


def test_print_warning(capsys: Any) -> None:
    """Test print_warning function."""
    helpers.print_warning("Test message")
    out = capsys.readouterr().out
    assert "Test message" in out


def test_print_error(capsys: Any) -> None:
    """Test print_error function."""
    helpers.print_error("Test message")
    out = capsys.readouterr().out
    assert "Test message" in out


def test_print_info(capsys: Any) -> None:
    """Test print_info function."""
    helpers.print_info("Test message")
    out = capsys.readouterr().out
    assert "Test message" in out


def test_validate_base32_edge_cases() -> None:
    """Test validate_base32 with edge cases."""
    # Test very long Base32 string
    long_secret = "JBSWY3DPEHPK3PXP" * 10
    assert helpers.validate_base32(long_secret) is True

    # Test Base32 with padding
    assert helpers.validate_base32("JBSWY3DPEHPK3PXP======") is True

    # Test Base32 with mixed case
    assert helpers.validate_base32("jbswy3dpehpk3pxp") is True

    # Test Base32 with spaces in middle
    assert helpers.validate_base32("JB SW Y3 DP EH PK 3P XP") is True


def test_parse_otpauth_url_edge_cases() -> None:
    """Test parse_otpauth_url with edge cases."""
    # Test with missing issuer parameter
    result = helpers.parse_otpauth_url(
        "otpauth://totp/GitHub:octocat?secret=JBSWY3DPEHPK3PXP"
    )
    assert result["issuer"] == "GitHub"
    assert result["label"] == "octocat"

    # Test with empty query parameters
    result = helpers.parse_otpauth_url(
        "otpauth://totp/GitHub:octocat?secret=JBSWY3DPEHPK3PXP&"
    )
    assert result["issuer"] == "GitHub"
    assert result["label"] == "octocat"
    assert result["secret"] == "JBSWY3DPEHPK3PXP"

    # Test with additional query parameters
    result = helpers.parse_otpauth_url(
        "otpauth://totp/GitHub:octocat?secret=JBSWY3DPEHPK3PXP&issuer=GitHub&algorithm=SHA1&digits=6&period=30"
    )
    assert result["issuer"] == "GitHub"
    assert result["label"] == "octocat"
    assert result["secret"] == "JBSWY3DPEHPK3PXP"


def test_print_header(capsys: Any) -> None:
    """Test print_header function."""
    helpers.print_header("Test message")
    out = capsys.readouterr().out
    assert "Test message" in out


def test_get_password_for_vault_password_file_read_error(
    tmp_path: Path, monkeypatch: Any
) -> None:
    """Test password file read error handling."""
    # Create a file that will cause read error
    password_file = tmp_path / "password.txt"
    password_file.write_text("test")

    # Mock open to raise an exception
    def mock_open(*args: Any, **kwargs: Any) -> None:
        raise PermissionError("Permission denied")

    monkeypatch.setattr("builtins.open", mock_open)

    fake_ctx = type(
        "FakeContext",
        (),
        {"obj": {"password_file": str(password_file), "interactive": True}},
    )()
    with pytest.raises(typer.Exit):
        helpers.get_password_for_vault(fake_ctx, new_vault=False)


def test_calculate_entropy() -> None:
    """Test password entropy calculation."""
    # Test passphrase (4 words: 11*4 = 44)
    assert helpers.calculate_entropy("correct horse battery staple") == 44

    # Test complex password
    assert helpers.calculate_entropy("P@ssw0rd123!") > 50

    # Test weak password
    assert helpers.calculate_entropy("password") < 40


def test_enforce_password_strength_weak_reject(monkeypatch: Any) -> None:
    """Test password strength enforcement with rejection."""

    # Mock load_config to return reject_weak=True and low min_entropy
    def mock_load_config() -> dict[str, Any]:
        return {
            "security": {"min_password_entropy": 100, "reject_weak_passwords": True}
        }

    monkeypatch.setattr("desktop_2fa.cli.helpers.load_config", mock_load_config)

    with pytest.raises(typer.Exit):
        helpers._enforce_password_strength("weak")


def test_enforce_password_strength_weak_warn(monkeypatch: Any) -> None:
    """Test password strength enforcement with warning."""

    # Mock load_config and typer.confirm
    def mock_load_config() -> dict[str, Any]:
        return {
            "security": {"min_password_entropy": 100, "reject_weak_passwords": False}
        }

    monkeypatch.setattr("desktop_2fa.cli.helpers.load_config", mock_load_config)
    monkeypatch.setattr("typer.confirm", lambda msg: True)

    # Should not raise
    helpers._enforce_password_strength("weak")


def test_enforce_password_strength_weak_warn_reject(monkeypatch: Any) -> None:
    """Test password strength enforcement with warning and user rejection."""

    def mock_load_config() -> dict[str, Any]:
        return {
            "security": {"min_password_entropy": 100, "reject_weak_passwords": False}
        }

    monkeypatch.setattr("desktop_2fa.cli.helpers.load_config", mock_load_config)
    monkeypatch.setattr("typer.confirm", lambda msg: False)

    with pytest.raises(typer.Exit):
        helpers._enforce_password_strength("weak")


def test_load_config_missing_file() -> None:
    """Test load_config with missing config file."""
    config = helpers.load_config()
    assert config == {}


def test_validate_base32_invalid_decode(monkeypatch: Any) -> None:
    """Test validate_base32 with invalid decode."""

    # Mock base64.b32decode to raise exception
    def mock_b32decode(data: Any) -> None:
        raise Exception("Invalid base32")

    monkeypatch.setattr("base64.b32decode", mock_b32decode)

    assert not helpers.validate_base32("INVALID")


def test_get_password_for_vault_enforce_strength(monkeypatch: Any) -> None:
    """Test password strength enforcement in get_password_for_vault."""
    # Mock to not skip password checks
    monkeypatch.setattr("os.getenv", lambda key: None)

    # Mock load_config to return low min_entropy and reject_weak=False
    def mock_load_config() -> dict[str, Any]:
        return {
            "security": {"min_password_entropy": 10, "reject_weak_passwords": False}
        }

    monkeypatch.setattr("desktop_2fa.cli.helpers.load_config", mock_load_config)
    monkeypatch.setattr("typer.confirm", lambda msg: True)

    fake_ctx = type("FakeContext", (), {"obj": {"interactive": True}})()

    # Mock prompts
    responses = ["weak", "weak"]
    monkeypatch.setattr("typer.prompt", lambda text, hide_input=False: responses.pop(0))

    # Should not raise (accepts weak password with confirmation)
    password = helpers.get_password_for_vault(fake_ctx, new_vault=True)
    assert password == "weak"


def test_print_prompt(capsys: Any) -> None:
    """Test print_prompt function."""
    helpers.print_prompt("Test message")
    out = capsys.readouterr().out
    assert "Test message" in out



"""CLI command implementations for Desktop 2FA."""

from __future__ import annotations

from pathlib import Path

import typer

import desktop_2fa.cli.helpers as helpers
from desktop_2fa.vault import Vault
from desktop_2fa.vault.vault import (
    CorruptedVault,
    InvalidPassword,
    UnsupportedFormat,
    VaultIOError,
)


def _path() -> Path:
    return Path(helpers.get_vault_path())


def add_entry_interactive(name: str, issuer: str, secret: str, ctx: typer.Context) -> None:
    """Add entry in interactive mode with explicit name."""
    path = _path()

    # Validate Base32 secret
    if not helpers.validate_base32(secret):
        helpers.print_error("Invalid secret: not valid Base32.")
        helpers.print_info("Example: ABCDEFGHIJKL2345")
        return

    if not path.exists():
        helpers.print_warning("No vault found.")
        helpers.print_info("A new encrypted vault will be created.")
        password = helpers.get_password_for_vault(ctx, new_vault=True)
        vault = Vault()
        vault.add_entry(name=name, issuer=issuer, secret=secret)
        vault.save(path, password)
        helpers.print_success("Vault created.")
        helpers.print_success(f"Entry added: {name}")
    else:
        password = helpers.get_password_for_vault(ctx, new_vault=False)
        try:
            vault = Vault.load(path, password)
            vault.add_entry(name=name, issuer=issuer, secret=secret)
            vault.save(path, password)
            helpers.print_success(f"Entry added: {name}")
        except ValueError as e:
            if "already exists" in str(e):
                helpers.print_error(str(e))
            else:
                helpers.print_error("Invalid vault password.")
        except InvalidPassword:
            helpers.print_error("Invalid vault password.")
        except CorruptedVault:
            helpers.print_error("Vault file is corrupted.")
        except UnsupportedFormat:
            helpers.print_error("Vault file format is unsupported.")
        except VaultIOError:
            helpers.print_error("Failed to access vault file.")


def list_entries(ctx: typer.Context) -> None:
    path = _path()
    interactive = ctx.obj.get("interactive", False)
    if not path.exists():
        if interactive:
            helpers.print_warning("No vault found.")
            helpers.print_info("A new encrypted vault will be created.")
        password = helpers.get_password_for_vault(ctx, new_vault=True)
        vault = Vault()
        vault.save(path, password)
        if interactive:
            helpers.print_success("Vault created.")
            helpers.print_info("No entries found.")
    else:
        password = helpers.get_password_for_vault(ctx, new_vault=False)
        try:
            vault = Vault.load(path, password)
        except InvalidPassword:
            if interactive:
                helpers.print_error("Invalid vault password.")
            return
        except CorruptedVault:
            if interactive:
                helpers.print_error("Vault file is corrupted.")
            return
        except UnsupportedFormat:
            if interactive:
                helpers.print_error("Vault file format is unsupported.")
            return
        except VaultIOError:
            if interactive:
                helpers.print_error("Failed to access vault file.")
            return
        if vault.entries:
            helpers.print_entries_table(vault.entries)
        else:
            if interactive:
                helpers.print_info("No entries found.")


def add_entry(name: str, issuer: str, secret: str, ctx: typer.Context) -> None:
    path = _path()

    # Parse otpauth URL if provided
    if issuer.startswith("otpauth://"):
        try:
            parsed = helpers.parse_otpauth_url(issuer)
            issuer = parsed["issuer"]
            name = parsed["label"]
            secret = parsed["secret"]
        except ValueError as e:
            helpers.print_error(f"Invalid otpauth URL: {e}")
            return

    # Validate Base32 secret
    if not helpers.validate_base32(secret):
        helpers.print_error("Invalid secret: not valid Base32.")
        helpers.print_info("Example: ABCDEFGHIJKL2345")
        return

    if not path.exists():
        helpers.print_warning("No vault found.")
        helpers.print_info("A new encrypted vault will be created.")
        password = helpers.get_password_for_vault(ctx, new_vault=True)
        vault = Vault()
        vault.add_entry(name=name, issuer=issuer, secret=secret)
        vault.save(path, password)
        helpers.print_success("Vault created.")
        helpers.print_success(f"Entry added: {name}")
    else:
        password = helpers.get_password_for_vault(ctx, new_vault=False)
        try:
            vault = Vault.load(path, password)
            vault.add_entry(name=name, issuer=issuer, secret=secret)
            vault.save(path, password)
            helpers.print_success(f"Entry added: {name}")
        except ValueError as e:
            if "already exists" in str(e):
                helpers.print_error(str(e))
            else:
                helpers.print_error("Invalid vault password.")
        except InvalidPassword:
            helpers.print_error("Invalid vault password.")
        except CorruptedVault:
            helpers.print_error("Vault file is corrupted.")
        except UnsupportedFormat:
            helpers.print_error("Vault file format is unsupported.")
        except VaultIOError:
            helpers.print_error("Failed to access vault file.")


def generate_code(name: str, ctx: typer.Context) -> None:
    path = _path()
    if not path.exists():
        helpers.print_warning("No vault found.")
        helpers.print_info("Nothing to generate.")
        return
    password = helpers.get_password_for_vault(ctx, new_vault=False)
    try:
        vault = Vault.load(path, password)
        entry = vault.get_entry(name)
        from desktop_2fa.totp.generator import generate

        code = generate(
            secret=entry.secret,
            digits=entry.digits,
            period=entry.period,
            algorithm=entry.algorithm,
        )
        print(code)
    except ValueError as e:
        if "not found" in str(e):
            raise
        helpers.print_error("Invalid vault password.")
    except InvalidPassword:
        helpers.print_error("Invalid vault password.")
    except CorruptedVault:
        helpers.print_error("Vault file is corrupted.")
    except UnsupportedFormat:
        helpers.print_error("Vault file format is unsupported.")
    except VaultIOError:
        helpers.print_error("Failed to access vault file.")


def remove_entry(name: str, ctx: typer.Context) -> None:
    path = _path()
    if not path.exists():
        helpers.print_warning("No vault found.")
        return
    password = helpers.get_password_for_vault(ctx, new_vault=False)
    try:
        vault = Vault.load(path, password)
        vault.remove_entry(name)
        vault.save(path, password)
        helpers.print_success(f"Removed entry: {name}")
    except ValueError as e:
        if "not found" in str(e):
            raise
        helpers.print_error("Invalid vault password.")
    except InvalidPassword:
        helpers.print_error("Invalid vault password.")
    except CorruptedVault:
        helpers.print_error("Vault file is corrupted.")
    except UnsupportedFormat:
        helpers.print_error("Vault file format is unsupported.")
    except VaultIOError:
        helpers.print_error("Failed to access vault file.")


def rename_entry(old: str, new: str, ctx: typer.Context) -> None:
    path = _path()
    if not path.exists():
        helpers.print_warning("No vault found.")
        return
    password = helpers.get_password_for_vault(ctx, new_vault=False)
    try:
        vault = Vault.load(path, password)
        entry = vault.get_entry(old)
        entry.account_name = new
        entry.issuer = new
        vault.save(path, password)
        helpers.print_success(f"Renamed '{old}' → '{new}'")
    except ValueError as e:
        if "not found" in str(e):
            raise
        helpers.print_error("Invalid vault password.")
    except InvalidPassword:
        helpers.print_error("Invalid vault password.")
    except CorruptedVault:
        helpers.print_error("Vault file is corrupted.")
    except UnsupportedFormat:
        helpers.print_error("Vault file format is unsupported.")
    except VaultIOError:
        helpers.print_error("Failed to access vault file.")


def export_vault(export_path: str, ctx: typer.Context) -> None:
    path = _path()
    if not path.exists():
        helpers.print_warning("No vault found.")
        return
    password = helpers.get_password_for_vault(ctx, new_vault=False)
    try:
        vault = Vault.load(path, password)
        vault.save(Path(export_path), password)
        helpers.print_success(f"Exported vault to: {export_path}")
    except InvalidPassword:
        helpers.print_error("Invalid vault password.")
    except CorruptedVault:
        helpers.print_error("Vault file is corrupted.")
    except UnsupportedFormat:
        helpers.print_error("Vault file format is unsupported.")
    except VaultIOError:
        helpers.print_error("Failed to access vault file.")


def import_vault(source: str, force: bool, ctx: typer.Context) -> None:
    path = _path()
    if path.exists() and not force:
        helpers.print_error(
            "Refusing to overwrite existing vault. Use --force to proceed."
        )
        raise typer.Exit(1)
    password = helpers.get_password_for_vault(ctx, new_vault=False)
    try:
        vault = Vault.load(Path(source), password)
        vault.save(path, password)
        helpers.print_success(f"Vault imported from {source}")
    except VaultIOError:
        raise
    except InvalidPassword:
        helpers.print_error("Invalid vault password.")
    except CorruptedVault:
        helpers.print_error("Source vault file is corrupted.")
    except UnsupportedFormat:
        helpers.print_error("Source vault file format is unsupported.")


def _get_backup_path(base_path: Path) -> Path:
    """Get the next available backup path with auto-suffixing."""
    backup_path = base_path.with_suffix(".backup.bin")
    if not backup_path.exists():
        return backup_path
    counter = 1
    while True:
        suffixed_path = base_path.with_suffix(f".backup-{counter}.bin")
        if not suffixed_path.exists():
            return suffixed_path
        counter += 1


def backup_vault(ctx: typer.Context) -> None:
    path = _path()
    if not path.exists():
        helpers.print_warning("No vault found.")
        return
    password = helpers.get_password_for_vault(ctx, new_vault=False)
    try:
        vault = Vault.load(path, password)
        backup_path = _get_backup_path(path)
        vault.save(backup_path, password)
        helpers.print_success(f"Backup created: {backup_path}")
    except InvalidPassword:
        helpers.print_error("Invalid vault password.")
    except CorruptedVault:
        helpers.print_error("Vault file is corrupted.")
    except UnsupportedFormat:
        helpers.print_error("Vault file format is unsupported.")
    except VaultIOError:
        helpers.print_error("Failed to access vault file.")


def init_vault(force: bool, ctx: typer.Context) -> None:
    """Initialize a new vault explicitly."""
    path = _path()
    if path.exists() and not force:
        helpers.print_info("Vault already exists.")
        helpers.print_info("Use --force to overwrite.")
        return

    if path.exists() and force:
        helpers.print_error("Existing vault will be overwritten.")

    password = helpers.get_password_for_vault(ctx, new_vault=True)
    vault = Vault()
    vault.save(path, password)
    helpers.print_success("Vault created.")



from pathlib import Path
from typing import Any

import pytest

from desktop_2fa.vault import Vault
from desktop_2fa.vault.models import TotpEntry, VaultData


def test_vault_roundtrip(tmp_path: Path) -> None:
    path = tmp_path / "vault.bin"

    vault = Vault()
    vault.add_entry("GitHub", "GitHub", "JBSWY3DPEHPK3PXP")
    vault.save(str(path))

    loaded = Vault.load(str(path))
    entry = loaded.get_entry("GitHub")
    assert entry.secret == "JBSWY3DPEHPK3PXP"


def test_vault_file_is_binary(tmp_path: Path) -> None:
    path = tmp_path / "vault.bin"

    vault = Vault()
    vault.add_entry("Test", "Test", "JBSWY3DPEHPK3PXP")
    vault.save(str(path))

    raw = path.read_bytes()
    assert raw != b""
    # upewniamy się, że jawne dane nie występują w pliku
    assert b"Test" not in raw
    assert b"JBSWY3DPEHPK3PXP" not in raw


def test_vault_tampering_detection(tmp_path: Path) -> None:
    path = tmp_path / "vault.bin"

    vault = Vault()
    vault.add_entry("X", "X", "JBSWY3DPEHPK3PXP")
    vault.save(str(path))

    raw = bytearray(path.read_bytes())
    raw[10] ^= 0xFF  # celowa korupcja
    path.write_bytes(raw)

    with pytest.raises(Exception):
        Vault.load(str(path))


def test_totp_entry_invalid_secret() -> None:
    with pytest.raises(ValueError, match="Invalid Base32 TOTP secret"):
        TotpEntry(issuer="Test", account_name="Test", secret="INVALID")


def test_totp_entry_invalid_period() -> None:
    with pytest.raises(ValueError, match="TOTP period must be positive"):
        TotpEntry(
            issuer="Test", account_name="Test", secret="JBSWY3DPEHPK3PXP", period=0
        )


def test_vault_load_invalid_magic_header(tmp_path: Path) -> None:
    """Test loading vault with invalid magic header."""
    path = tmp_path / "vault.bin"

    # Create a file with invalid magic header
    invalid_data = b"WRNG" + b"\x01" + b"16byte_salt_here" + b"encrypted_data"
    path.write_bytes(invalid_data)

    with pytest.raises(
        Exception, match="Invalid vault file format: incorrect magic header"
    ):
        Vault.load(str(path))


def test_vault_load_unsupported_version(tmp_path: Path) -> None:
    """Test loading vault with unsupported version."""
    path = tmp_path / "vault.bin"

    # Create a file with unsupported version
    invalid_data = b"D2FA" + b"\x02" + b"16byte_salt_here" + b"encrypted_data"
    path.write_bytes(invalid_data)

    with pytest.raises(Exception, match="Unsupported vault file version"):
        Vault.load(str(path))


def test_vault_load_empty_encrypted_blob(tmp_path: Path) -> None:
    """Test loading vault with empty encrypted blob."""
    path = tmp_path / "vault.bin"

    # Create a file with empty encrypted blob
    invalid_data = b"D2FA" + b"\x01" + b"16byte_salt_here"
    path.write_bytes(invalid_data)

    with pytest.raises(Exception, match="Vault file is invalid: empty encrypted blob"):
        Vault.load(str(path))


def test_vault_load_corrupted_json_data(tmp_path: Path) -> None:
    """Test loading vault with corrupted JSON data."""
    path = tmp_path / "vault.bin"

    # Create a valid vault structure but with corrupted JSON
    from desktop_2fa.crypto.aesgcm import encrypt
    from desktop_2fa.crypto.argon2 import derive_key

    password = "test_password"
    salt = b"16byte_salt_here"
    key = derive_key(password, salt)

    # Corrupted JSON data
    corrupted_json = b'{"invalid": json}'
    encrypted = encrypt(key, corrupted_json)

    valid_data = b"D2FA" + b"\x01" + salt + encrypted
    path.write_bytes(valid_data)

    with pytest.raises(Exception, match="Invalid password or corrupted vault"):
        Vault.load(str(path))


def test_vault_save_io_error(tmp_path: Path, monkeypatch: Any) -> None:
    """Test saving vault with IO error."""
    path = tmp_path / "vault.bin"

    vault = Vault()
    vault.add_entry("Test", "Test", "JBSWY3DPEHPK3PXP")

    # Mock os.open to raise an exception
    def mock_open(*args: Any, **kwargs: Any) -> None:
        raise OSError("Permission denied")

    monkeypatch.setattr("os.open", mock_open)

    with pytest.raises(Exception, match="Failed to save vault: Permission denied"):
        vault.save(str(path))


def test_vault_save_cleanup_on_error(tmp_path: Path, monkeypatch: Any) -> None:
    """Test that temporary file is cleaned up when save fails."""
    path = tmp_path / "vault.bin"
    temp_path = path.with_suffix(".tmp")

    vault = Vault()
    vault.add_entry("Test", "Test", "JBSWY3DPEHPK3PXP")

    # Mock os.open to succeed but os.fdopen to fail
    def mock_open(*args: Any, **kwargs: Any) -> int:
        return 1  # fake file descriptor

    def mock_fdopen(*args: Any, **kwargs: Any) -> None:
        raise OSError("Write error")

    monkeypatch.setattr("os.open", mock_open)
    monkeypatch.setattr("os.fdopen", mock_fdopen)

    with pytest.raises(Exception, match="Failed to save vault: Write error"):
        vault.save(str(path))

    # Temporary file should be cleaned up
    assert not temp_path.exists()


def test_vault_load_malformed_json_edge_cases(tmp_path: Path) -> None:
    """Test loading vault with malformed JSON edge cases."""
    path = tmp_path / "vault.bin"

    # Create a valid vault structure but with malformed JSON that passes basic validation
    from desktop_2fa.crypto.aesgcm import encrypt
    from desktop_2fa.crypto.argon2 import derive_key

    password = "test_password"
    salt = b"16byte_salt_here"
    key = derive_key(password, salt)

    # JSON that's valid but has invalid structure for VaultData
    malformed_json = b'{"version": "invalid_type"}'
    encrypted = encrypt(key, malformed_json)

    valid_data = b"D2FA" + b"\x01" + salt + encrypted
    path.write_bytes(valid_data)

    with pytest.raises(Exception, match="Vault contains invalid data"):
        Vault.load(str(path), password="test_password")


def test_vault_save_os_level_errors(tmp_path: Path, monkeypatch: Any) -> None:
    """Test saving vault with OS-level file errors."""
    path = tmp_path / "vault.bin"

    vault = Vault()
    vault.add_entry("Test", "Test", "JBSWY3DPEHPK3PXP")

    # Mock os.open to raise a specific OS error
    def mock_open(*args: Any, **kwargs: Any) -> None:
        raise OSError(28, "No space left on device")

    monkeypatch.setattr("os.open", mock_open)

    with pytest.raises(
        Exception, match="Failed to save vault: No space left on device"
    ):
        vault.save(str(path))


def test_vault_save_cleanup_on_os_error(tmp_path: Path, monkeypatch: Any) -> None:
    """Test that temporary file is cleaned up when OS-level save fails."""
    path = tmp_path / "vault.bin"
    temp_path = path.with_suffix(".tmp")

    vault = Vault()
    vault.add_entry("Test", "Test", "JBSWY3DPEHPK3PXP")

    # Mock os.open to succeed but os.fsync to fail with OS error
    def mock_open(*args: Any, **kwargs: Any) -> int:
        return 1  # fake file descriptor

    def mock_fsync(fd: int) -> None:
        raise OSError(28, "No space left on device")

    monkeypatch.setattr("os.open", mock_open)
    monkeypatch.setattr("os.fsync", mock_fsync)

    with pytest.raises(Exception, match="Failed to save vault:"):
        vault.save(str(path))

    # Temporary file should be cleaned up
    assert not temp_path.exists()


def test_vault_load_json_validation_edge_cases(tmp_path: Path) -> None:
    """Test loading vault with JSON that passes validation but fails model validation."""
    path = tmp_path / "vault.bin"

    # Create a valid vault structure but with JSON that has invalid field types
    from desktop_2fa.crypto.aesgcm import encrypt
    from desktop_2fa.crypto.argon2 import derive_key

    password = "test_password"
    salt = b"16byte_salt_here"
    key = derive_key(password, salt)

    # JSON with invalid field types that will fail Pydantic validation
    invalid_json = b'{"entries": [{"issuer": 123, "account_name": "test", "secret": "JBSWY3DPEHPK3PXP"}]}'
    encrypted = encrypt(key, invalid_json)

    valid_data = b"D2FA" + b"\x01" + salt + encrypted
    path.write_bytes(valid_data)

    with pytest.raises(Exception, match="Vault contains invalid data"):
        Vault.load(str(path), password="test_password")


def test_vault_load_no_password_default(tmp_path: Path) -> None:
    """Test loading vault with no password (uses default empty string)."""
    path = tmp_path / "vault.bin"

    vault = Vault()
    vault.add_entry("Test", "Test", "JBSWY3DPEHPK3PXP")
    vault.save(str(path), password=None)  # Save with no password

    # Load with no password (should use default empty string)
    loaded = Vault.load(str(path), password=None)
    entry = loaded.get_entry("Test")
    assert entry.secret == "JBSWY3DPEHPK3PXP"


def test_vault_save_no_password_default(tmp_path: Path) -> None:
    """Test saving vault with no password (uses default empty string)."""
    path = tmp_path / "vault.bin"

    vault = Vault()
    vault.add_entry("Test", "Test", "JBSWY3DPEHPK3PXP")

    # Save with no password (should use default empty string)
    vault.save(str(path), password=None)

    # File should be created
    assert path.exists()

    # Should be able to load with no password
    loaded = Vault.load(str(path), password=None)
    entry = loaded.get_entry("Test")
    assert entry.secret == "JBSWY3DPEHPK3PXP"


def test_vault_load_file_not_found(tmp_path: Path) -> None:
    """Test loading vault from non-existent file."""
    path = tmp_path / "nonexistent.bin"

    with pytest.raises(Exception, match="Failed to read vault file"):
        Vault.load(str(path))


def test_vault_load_file_too_short(tmp_path: Path) -> None:
    """Test loading vault file that is too short."""
    path = tmp_path / "vault.bin"

    # Create a file that's too short
    short_data = b"D2FA\x01"
    path.write_bytes(short_data)

    with pytest.raises(Exception, match="Vault file is too short or invalid format"):
        Vault.load(str(path))


def test_vault_add_entry_duplicate_name() -> None:
    """Test that adding an entry with a duplicate name raises ValueError."""
    vault = Vault()
    vault.add_entry("GitHub", "GitHub", "JBSWY3DPEHPK3PXP")

    with pytest.raises(ValueError, match='An entry with name "GitHub" already exists'):
        vault.add_entry("GitHub", "GitHub", "ABCDEFGHIJKL1234")


def test_vault_load_duplicate_names_warning(tmp_path: Path, capsys: Any) -> None:
    """Test that loading a vault with duplicate names shows a warning but doesn't fail."""
    path = tmp_path / "vault.bin"

    # Create vault data with duplicate names manually
    data = VaultData(entries=[
        TotpEntry(account_name="GitHub", issuer="GitHub", secret="JBSWY3DPEHPK3PXP"),
        TotpEntry(account_name="GitHub", issuer="GitHub", secret="JBSWY3DPEHPK3PXP"),
    ])
    vault = Vault(data)
    vault.save(str(path))

    # Load the vault - should show warning but succeed
    loaded = Vault.load(str(path))

    # Check that warning was printed
    captured = capsys.readouterr()
    assert 'Warning: Your vault contains multiple entries with the same name: "GitHub"' in captured.out

    # Vault should still load successfully
    assert len(loaded.entries) == 2



"""Pydantic models for vault data structures."""

import base64
from typing import Literal, Optional

from pydantic import BaseModel, Field, field_validator


class TotpEntry(BaseModel):
    """Model for a TOTP entry in the vault."""

    issuer: Optional[str] = Field(None)
    account_name: Optional[str] = Field(None)
    secret: str = Field(...)
    digits: Literal[6, 7, 8] = 6
    period: int = 30
    algorithm: Literal["SHA1", "SHA256", "SHA512"] = "SHA1"

    @field_validator("secret")
    @classmethod
    def validate_base32_secret(cls, v: str) -> str:
        """Validate that the secret is valid Base32.

        Args:
            v: The secret value to validate.

        Returns:
            The validated secret.

        Raises:
            ValueError: If the secret is not valid Base32.
        """
        try:
            base64.b32decode(v.upper())
        except Exception:
            raise ValueError("Invalid Base32 TOTP secret")
        return v

    @field_validator("period")
    @classmethod
    def validate_period(cls, v: int) -> int:
        """Validate that the period is positive.

        Args:
            v: The period value to validate.

        Returns:
            The validated period.

        Raises:
            ValueError: If the period is not positive.
        """
        if v <= 0:
            raise ValueError("TOTP period must be positive")
        return v


class VaultData(BaseModel):
    """Model for the vault data structure."""

    version: int = 1
    entries: list[TotpEntry] = Field(default_factory=list)



import shutil
from pathlib import Path
from typing import Any

import pytest
import typer

from desktop_2fa.cli import commands, helpers
from desktop_2fa.vault.vault import VaultIOError

TEST_PASSWORD = "jawislajawisla"


@pytest.fixture
def fake_vault_env(tmp_path: Path, monkeypatch: Any) -> Path:
    fake_vault = tmp_path / "vault"

    monkeypatch.setattr(
        "desktop_2fa.cli.helpers.get_vault_path",
        lambda: str(fake_vault),
    )

    if fake_vault.parent.exists():
        shutil.rmtree(fake_vault.parent)
    fake_vault.parent.mkdir(parents=True, exist_ok=True)

    return fake_vault


@pytest.fixture
def fake_ctx() -> Any:
    class FakeContext:
        def __init__(self) -> None:
            self.obj: dict[str, Any] = {"interactive": True, "password": TEST_PASSWORD}

    return FakeContext()


@pytest.fixture
def fake_ctx_wrong_password() -> Any:
    class FakeContext:
        def __init__(self) -> None:
            self.obj: dict[str, Any] = {
                "interactive": True,
                "password": "wrongpassword",
            }

    return FakeContext()


def test_list_entries_empty(fake_vault_env: Path, capsys: Any, fake_ctx: Any) -> None:
    commands.list_entries(fake_ctx)
    out = capsys.readouterr().out.strip().splitlines()
    assert out == [
        "No vault found.",
        "A new encrypted vault will be created.",
        "Vault created.",
        "No entries found.",
    ]


def test_add_entry_and_list(fake_vault_env: Path, capsys: Any, fake_ctx: Any) -> None:
    commands.add_entry("GitHub", "GitHub", "JBSWY3DPEHPK3PXP", fake_ctx)

    # po add_entry:
    out = capsys.readouterr().out.strip().splitlines()
    assert out == [
        "No vault found.",
        "A new encrypted vault will be created.",
        "Vault created.",
        "Entry added: GitHub",
    ]

    commands.list_entries(fake_ctx)
    out = capsys.readouterr().out.strip().splitlines()
    assert out == ["- GitHub (GitHub)"]

    vault = helpers.load_vault(fake_vault_env, TEST_PASSWORD)
    assert len(vault.entries) == 1
    assert vault.entries[0].issuer == "GitHub"
    assert vault.entries[0].account_name == "GitHub"
    assert vault.entries[0].secret == "JBSWY3DPEHPK3PXP"


def test_generate_code(fake_vault_env: Path, capsys: Any, fake_ctx: Any) -> None:
    commands.add_entry("GitHub", "GitHub", "JBSWY3DPEHPK3PXP", fake_ctx)
    capsys.readouterr()  # czyścimy output po add_entry

    commands.generate_code("GitHub", fake_ctx)
    out = capsys.readouterr().out.strip()

    # generate_code wypisuje tylko kod, jedną linię
    lines = out.splitlines()
    code = lines[-1]
    assert len(code) == 6
    assert code.isdigit()


def test_generate_code_missing_entry_raises(
    fake_vault_env: Path, fake_ctx: Any
) -> None:
    # Create empty vault
    from desktop_2fa.vault import Vault

    vault = Vault()
    vault.save(fake_vault_env, TEST_PASSWORD)
    with pytest.raises(ValueError):
        commands.generate_code("Nope", fake_ctx)


def test_remove_entry(fake_vault_env: Path, fake_ctx: Any) -> None:
    commands.add_entry("GitHub", "GitHub", "JBSWY3DPEHPK3PXP", fake_ctx)
    commands.remove_entry("GitHub", fake_ctx)

    vault = helpers.load_vault(fake_vault_env, TEST_PASSWORD)
    assert len(vault.entries) == 0


def test_remove_entry_missing_raises(fake_vault_env: Path, fake_ctx: Any) -> None:
    # Create empty vault
    from desktop_2fa.vault import Vault

    vault = Vault()
    vault.save(fake_vault_env, TEST_PASSWORD)
    with pytest.raises(ValueError):
        commands.remove_entry("Nope", fake_ctx)


def test_rename_entry(fake_vault_env: Path, capsys: Any, fake_ctx: Any) -> None:
    commands.add_entry("GitHub", "GitHub", "JBSWY3DPEHPK3PXP", fake_ctx)
    capsys.readouterr()  # czyścimy output po add_entry

    commands.rename_entry("GitHub", "NewGitHub", fake_ctx)

    vault = helpers.load_vault(fake_vault_env, TEST_PASSWORD)
    assert len(vault.entries) == 1
    entry = vault.entries[0]
    assert entry.account_name == "NewGitHub"
    assert entry.issuer == "NewGitHub"


def test_rename_entry_missing_raises(fake_vault_env: Path, fake_ctx: Any) -> None:
    # Create empty vault
    from desktop_2fa.vault import Vault

    vault = Vault()
    vault.save(fake_vault_env, TEST_PASSWORD)
    with pytest.raises(ValueError):
        commands.rename_entry("Old", "New", fake_ctx)


def test_export_vault(
    fake_vault_env: Path, tmp_path: Path, capsys: Any, fake_ctx: Any
) -> None:
    commands.add_entry("GitHub", "GitHub", "JBSWY3DPEHPK3PXP", fake_ctx)
    capsys.readouterr()  # wyczyść output po add_entry

    export_path = tmp_path / "export.bin"
    commands.export_vault(str(export_path), fake_ctx)

    assert export_path.exists()
    assert export_path.stat().st_size > 0

    out = capsys.readouterr().out
    assert "Exported vault to:" in out


def test_export_vault_missing_file(
    fake_vault_env: Path, tmp_path: Path, monkeypatch: Any, capsys: Any, fake_ctx: Any
) -> None:
    fake_vault = tmp_path / "vault_missing"
    monkeypatch.setattr(
        "desktop_2fa.cli.helpers.get_vault_path",
        lambda: str(fake_vault),
    )

    commands.export_vault(str(tmp_path / "any.bin"), fake_ctx)
    out = capsys.readouterr().out
    # zgodnie z aktualnym helpers.export_vault, brak explicit checka,
    # więc tutaj nie wymuszamy konkretnego komunikatu – tylko, że coś wypisuje.
    assert "No vault found." in out


def test_import_vault(
    fake_vault_env: Path, tmp_path: Path, capsys: Any, fake_ctx: Any
) -> None:
    src = tmp_path / "src.bin"

    commands.add_entry("GitHub", "GitHub", "JBSWY3DPEHPK3PXP", fake_ctx)
    capsys.readouterr()  # output po add_entry

    commands.export_vault(str(src), fake_ctx)
    capsys.readouterr()  # output po export

    vault = helpers.load_vault(fake_vault_env, TEST_PASSWORD)
    vault.entries.clear()
    vault.save(fake_vault_env, TEST_PASSWORD)

    commands.import_vault(str(src), True, fake_ctx)

    vault = helpers.load_vault(fake_vault_env, TEST_PASSWORD)
    assert len(vault.entries) == 1
    assert vault.entries[0].issuer == "GitHub"
    assert vault.entries[0].secret == "JBSWY3DPEHPK3PXP"

    out = capsys.readouterr().out
    assert "Vault imported from" in out


def test_import_vault_missing_source(
    fake_vault_env: Path, tmp_path: Path, capsys: Any, fake_ctx: Any
) -> None:
    missing = tmp_path / "nope.bin"
    with pytest.raises(VaultIOError):
        commands.import_vault(str(missing), False, fake_ctx)


def test_import_vault_refuses_overwrite_without_force(
    fake_vault_env: Path, tmp_path: Path, capsys: Any, fake_ctx: Any
) -> None:
    # Create existing vault
    commands.add_entry("GitHub", "GitHub", "JBSWY3DPEHPK3PXP", fake_ctx)
    capsys.readouterr()  # clear output

    # Create source vault
    src = tmp_path / "src.bin"
    from desktop_2fa.vault import Vault

    vault = Vault()
    vault.save(src, TEST_PASSWORD)

    # Try import without force - should refuse
    with pytest.raises(typer.Exit):
        commands.import_vault(str(src), False, fake_ctx)

    out = capsys.readouterr().out.strip()
    assert out == "Refusing to overwrite existing vault. Use --force to proceed."


def test_backup_vault(fake_vault_env: Path, capsys: Any, fake_ctx: Any) -> None:
    commands.add_entry("GitHub", "GitHub", "JBSWY3DPEHPK3PXP", fake_ctx)
    capsys.readouterr()  # output po add_entry

    # First backup creates backup.bin
    backup_path = fake_vault_env.with_suffix(".backup.bin")
    commands.backup_vault(fake_ctx)

    assert backup_path.exists()
    assert backup_path.stat().st_size > 0

    out = capsys.readouterr().out
    assert "Backup created:" in out

    # Second backup creates backup-1.bin
    commands.backup_vault(fake_ctx)
    backup_path_1 = fake_vault_env.with_suffix(".backup-1.bin")

    assert backup_path_1.exists()
    assert backup_path_1.stat().st_size > 0

    out = capsys.readouterr().out
    assert "Backup created:" in out


def test_backup_vault_missing(
    fake_vault_env: Path, capsys: Any, monkeypatch: Any, tmp_path: Path, fake_ctx: Any
) -> None:
    fake_missing = tmp_path / "no_vault_here"
    monkeypatch.setattr(
        "desktop_2fa.cli.helpers.get_vault_path",
        lambda: str(fake_missing),
    )

    commands.backup_vault(fake_ctx)
    out = capsys.readouterr().out
    # Now prints "No vault found."
    assert "No vault found." in out


def test_generate_code_invalid_password(
    fake_vault_env: Path, capsys: Any, fake_ctx_wrong_password: Any
) -> None:
    # Create vault with correct password
    from desktop_2fa.vault import Vault

    vault = Vault()
    vault.save(fake_vault_env, TEST_PASSWORD)

    commands.generate_code("nonexistent", fake_ctx_wrong_password)
    out = capsys.readouterr().out.strip()
    assert out == "Invalid vault password."


def test_add_entry_invalid_password(
    fake_vault_env: Path, capsys: Any, fake_ctx_wrong_password: Any
) -> None:
    # Create vault with correct password
    from desktop_2fa.vault import Vault

    vault = Vault()
    vault.save(fake_vault_env, TEST_PASSWORD)

    commands.add_entry("Test", "Test", "JBSWY3DPEHPK3PXP", fake_ctx_wrong_password)
    out = capsys.readouterr().out.strip()
    assert out == "Invalid vault password."


def test_list_entries_invalid_password(
    fake_vault_env: Path, capsys: Any, fake_ctx_wrong_password: Any
) -> None:
    # Create vault with correct password
    from desktop_2fa.vault import Vault

    vault = Vault()
    vault.save(fake_vault_env, TEST_PASSWORD)

    commands.list_entries(fake_ctx_wrong_password)
    out = capsys.readouterr().out.strip()
    assert out == "Invalid vault password."


def test_remove_entry_invalid_password(
    fake_vault_env: Path, capsys: Any, fake_ctx_wrong_password: Any
) -> None:
    # Create vault with correct password
    from desktop_2fa.vault import Vault

    vault = Vault()
    vault.save(fake_vault_env, TEST_PASSWORD)

    commands.remove_entry("Test", fake_ctx_wrong_password)
    out = capsys.readouterr().out.strip()
    assert out == "Invalid vault password."


def test_export_vault_invalid_password(
    fake_vault_env: Path, tmp_path: Path, capsys: Any, fake_ctx_wrong_password: Any
) -> None:
    # Create vault with correct password
    from desktop_2fa.vault import Vault

    vault = Vault()
    vault.save(fake_vault_env, TEST_PASSWORD)

    export_path = tmp_path / "export.bin"
    commands.export_vault(str(export_path), fake_ctx_wrong_password)
    out = capsys.readouterr().out.strip()
    assert out == "Invalid vault password."


def test_import_vault_invalid_password(
    fake_vault_env: Path, tmp_path: Path, capsys: Any, fake_ctx_wrong_password: Any
) -> None:
    # Create source vault with correct password
    from desktop_2fa.vault import Vault

    vault = Vault()
    src = tmp_path / "src.bin"
    vault.save(src, TEST_PASSWORD)

    commands.import_vault(str(src), False, fake_ctx_wrong_password)
    out = capsys.readouterr().out.strip()
    assert out == "Invalid vault password."


def test_backup_vault_invalid_password(
    fake_vault_env: Path, capsys: Any, fake_ctx_wrong_password: Any
) -> None:
    # Create vault with correct password
    from desktop_2fa.vault import Vault

    vault = Vault()
    vault.save(fake_vault_env, TEST_PASSWORD)

    commands.backup_vault(fake_ctx_wrong_password)
    out = capsys.readouterr().out.strip()
    assert out == "Invalid vault password."


def test_add_entry_otpauth_url(
    fake_vault_env: Path, capsys: Any, fake_ctx: Any
) -> None:
    """Test adding entry from otpauth URL."""
    otpauth_url = "otpauth://totp/GitHub:octocat?secret=JBSWY3DPEHPK3PXP&issuer=GitHub"
    commands.add_entry("GitHub", otpauth_url, "", fake_ctx)

    out = capsys.readouterr().out.strip().splitlines()
    assert "Entry added: octocat" in out


def test_add_entry_invalid_otpauth_url(
    fake_vault_env: Path, capsys: Any, fake_ctx: Any
) -> None:
    """Test adding entry with invalid otpauth URL."""
    invalid_url = "otpauth://invalid"
    commands.add_entry("Test", invalid_url, "", fake_ctx)

    out = capsys.readouterr().out.strip()
    assert "Invalid otpauth URL:" in out


def test_add_entry_invalid_secret(
    fake_vault_env: Path, capsys: Any, fake_ctx: Any
) -> None:
    """Test adding entry with invalid Base32 secret."""
    commands.add_entry("Test", "Test", "invalid_secret", fake_ctx)

    out = capsys.readouterr().out.strip()
    assert "Invalid secret: not valid Base32." in out


def test_list_entries_existing_vault_no_entries(
    fake_vault_env: Path, capsys: Any, fake_ctx: Any
) -> None:
    """Test listing entries when vault exists but has no entries."""
    from desktop_2fa.vault import Vault

    vault = Vault()
    vault.save(fake_vault_env, TEST_PASSWORD)

    commands.list_entries(fake_ctx)

    out = capsys.readouterr().out.strip().splitlines()
    assert "No entries found." in out


def test_generate_code_existing_vault_no_entries(
    fake_vault_env: Path, capsys: Any, fake_ctx: Any
) -> None:
    """Test generating code when vault exists but has no entries."""
    from desktop_2fa.vault import Vault

    vault = Vault()
    vault.save(fake_vault_env, TEST_PASSWORD)

    with pytest.raises(ValueError, match="not found"):
        commands.generate_code("nonexistent", fake_ctx)


def test_remove_entry_existing_vault_no_entries(
    fake_vault_env: Path, capsys: Any, fake_ctx: Any
) -> None:
    """Test removing entry when vault exists but has no entries."""
    from desktop_2fa.vault import Vault

    vault = Vault()
    vault.save(fake_vault_env, TEST_PASSWORD)

    with pytest.raises(ValueError, match="not found"):
        commands.remove_entry("nonexistent", fake_ctx)


def test_rename_entry_existing_vault_no_entries(
    fake_vault_env: Path, capsys: Any, fake_ctx: Any
) -> None:
    """Test renaming entry when vault exists but has no entries."""
    from desktop_2fa.vault import Vault

    vault = Vault()
    vault.save(fake_vault_env, TEST_PASSWORD)

    with pytest.raises(ValueError, match="not found"):
        commands.rename_entry("old", "new", fake_ctx)


def test_export_vault_existing_vault_no_entries(
    fake_vault_env: Path, tmp_path: Path, capsys: Any, fake_ctx: Any
) -> None:
    """Test exporting vault when it exists but has no entries."""
    from desktop_2fa.vault import Vault

    vault = Vault()
    vault.save(fake_vault_env, TEST_PASSWORD)

    export_path = tmp_path / "export.bin"
    commands.export_vault(str(export_path), fake_ctx)

    assert export_path.exists()
    out = capsys.readouterr().out
    assert "Exported vault to:" in out


def test_import_vault_corrupted_source(
    fake_vault_env: Path, tmp_path: Path, capsys: Any, fake_ctx: Any
) -> None:
    """Test importing from corrupted vault file."""
    corrupted = tmp_path / "corrupted.bin"
    corrupted.write_text("corrupted data")

    commands.import_vault(str(corrupted), True, fake_ctx)
    out = capsys.readouterr().out.strip()
    assert "Source vault file format is unsupported." in out


def test_import_vault_unsupported_format(
    fake_vault_env: Path, tmp_path: Path, capsys: Any, fake_ctx: Any
) -> None:
    """Test importing from unsupported format."""
    unsupported = tmp_path / "unsupported.bin"
    # Create a file that looks like a vault but has unsupported format
    unsupported.write_bytes(b"unsupported_format_data")

    commands.import_vault(str(unsupported), True, fake_ctx)

    out = capsys.readouterr().out.strip()
    assert "Source vault file format is unsupported." in out


def test_backup_vault_existing_vault_no_entries(
    fake_vault_env: Path, capsys: Any, fake_ctx: Any
) -> None:
    """Test backing up vault when it exists but has no entries."""
    from desktop_2fa.vault import Vault

    vault = Vault()
    vault.save(fake_vault_env, TEST_PASSWORD)

    commands.backup_vault(fake_ctx)

    backup_path = fake_vault_env.with_suffix(".backup.bin")
    assert backup_path.exists()
    out = capsys.readouterr().out
    assert "Backup created:" in out


def test_init_vault_existing_no_force(
    fake_vault_env: Path, capsys: Any, fake_ctx: Any
) -> None:
    """Test initializing vault when it already exists without force."""
    from desktop_2fa.vault import Vault

    vault = Vault()
    vault.save(fake_vault_env, TEST_PASSWORD)

    commands.init_vault(False, fake_ctx)

    out = capsys.readouterr().out.strip()
    assert "Vault already exists." in out
    assert "Use --force to overwrite." in out


def test_init_vault_existing_with_force(
    fake_vault_env: Path, capsys: Any, fake_ctx: Any
) -> None:
    """Test initializing vault when it already exists with force."""
    from desktop_2fa.vault import Vault

    vault = Vault()
    vault.save(fake_vault_env, TEST_PASSWORD)

    commands.init_vault(True, fake_ctx)

    out = capsys.readouterr().out.strip()
    assert "Vault created." in out


def test_init_vault_new(fake_vault_env: Path, capsys: Any, fake_ctx: Any) -> None:
    """Test initializing new vault."""
    commands.init_vault(False, fake_ctx)

    out = capsys.readouterr().out.strip()
    assert "Vault created." in out
    assert fake_vault_env.exists()


def test_remove_entry_unsupported_format(
    fake_vault_env: Path, capsys: Any, fake_ctx: Any
) -> None:
    """Test removing entry from unsupported vault format."""
    # Create a file with wrong magic header
    fake_vault_env.write_bytes(
        b"WRNG" + b"\x01" + b"16byte_salt_here" + b"encrypted_data"
    )

    commands.remove_entry("GitHub", fake_ctx)

    out = capsys.readouterr().out.strip()
    assert "Vault file format is unsupported." in out


def test_rename_entry_unsupported_format(
    fake_vault_env: Path, capsys: Any, fake_ctx: Any
) -> None:
    """Test renaming entry in unsupported vault format."""
    # Create a file with wrong magic header
    fake_vault_env.write_bytes(
        b"WRNG" + b"\x01" + b"16byte_salt_here" + b"encrypted_data"
    )

    commands.rename_entry("GitHub", "NewGitHub", fake_ctx)

    out = capsys.readouterr().out.strip()
    assert "Vault file format is unsupported." in out


def test_export_vault_unsupported_format(
    fake_vault_env: Path, tmp_path: Path, capsys: Any, fake_ctx: Any
) -> None:
    """Test exporting from unsupported vault format."""
    # Create a file with wrong magic header
    fake_vault_env.write_bytes(
        b"WRNG" + b"\x01" + b"16byte_salt_here" + b"encrypted_data"
    )

    export_path = tmp_path / "export.bin"
    commands.export_vault(str(export_path), fake_ctx)

    out = capsys.readouterr().out.strip()
    assert "Vault file format is unsupported." in out


def test_backup_vault_unsupported_format(
    fake_vault_env: Path, capsys: Any, fake_ctx: Any
) -> None:
    """Test backing up unsupported vault format."""
    # Create a file with wrong magic header
    fake_vault_env.write_bytes(
        b"WRNG" + b"\x01" + b"16byte_salt_here" + b"encrypted_data"
    )

    commands.backup_vault(fake_ctx)

    out = capsys.readouterr().out.strip()
    assert "Vault file format is unsupported." in out


def test_add_entry_unsupported_format(
    fake_vault_env: Path, capsys: Any, fake_ctx: Any
) -> None:
    """Test adding entry to unsupported vault format."""
    # Create a file with wrong magic header
    fake_vault_env.write_bytes(
        b"WRNG" + b"\x01" + b"16byte_salt_here" + b"encrypted_data"
    )

    commands.add_entry("GitHub", "GitHub", "JBSWY3DPEHPK3PXP", fake_ctx)

    out = capsys.readouterr().out.strip()
    assert "Vault file format is unsupported." in out


def test_list_entries_unsupported_format(
    fake_vault_env: Path, capsys: Any, fake_ctx: Any
) -> None:
    """Test listing entries from unsupported vault format."""
    # Create a file with wrong magic header
    fake_vault_env.write_bytes(
        b"WRNG" + b"\x01" + b"16byte_salt_here" + b"encrypted_data"
    )

    commands.list_entries(fake_ctx)

    out = capsys.readouterr().out.strip()
    assert "Vault file format is unsupported." in out


def test_generate_code_unsupported_format(
    fake_vault_env: Path, capsys: Any, fake_ctx: Any
) -> None:
    """Test generating code from unsupported vault format."""
    # Create a file with wrong magic header
    fake_vault_env.write_bytes(
        b"WRNG" + b"\x01" + b"16byte_salt_here" + b"encrypted_data"
    )

    commands.generate_code("GitHub", fake_ctx)

    out = capsys.readouterr().out.strip()
    assert "Vault file format is unsupported." in out





"""AES-GCM encryption and decryption utilities."""

import os

from cryptography.hazmat.primitives.ciphers.aead import AESGCM


def encrypt(key: bytes, data: bytes) -> bytes:
    """Encrypt data using AES-GCM.

    Args:
        key: The encryption key (32 bytes for AES-256).
        data: The plaintext data to encrypt.

    Returns:
        The encrypted data in the format: 12-byte nonce + ciphertext + 16-byte authentication tag.
    """
    aes = AESGCM(key)
    nonce = os.urandom(12)
    return nonce + aes.encrypt(nonce, data, None)


def decrypt(key: bytes, blob: bytes) -> bytes:
    """Decrypt data using AES-GCM.

    Args:
        key: The decryption key (32 bytes for AES-256).
        blob: The encrypted data in the format: 12-byte nonce + ciphertext + 16-byte authentication tag.

    Returns:
        The decrypted plaintext data.

    Raises:
        ValueError: If decryption fails.
    """
    if len(blob) < 12:
        raise ValueError("Encrypted blob too short")
    nonce, ciphertext = blob[:12], blob[12:]
    aes = AESGCM(key)
    try:
        return aes.decrypt(nonce, ciphertext, None)
    except Exception as e:
        raise ValueError(f"Decryption failed: {e}") from e



"""Argon2 key derivation utilities."""

from argon2.low_level import Type, hash_secret_raw


def derive_key(password: str, salt: bytes) -> bytes:
    """Derive a key from password and salt using Argon2id.

    Argon2id parameters (version 1.3, stable for 2025-era hardware):
    - time_cost: 4 iterations (balances security and performance)
    - memory_cost: 131072 KiB (128 MiB) (resistant to GPU attacks)
    - parallelism: 2 threads (suitable for most systems)
    - hash_len: 32 bytes (for AES-256)
    - salt_len: 16 bytes (recommended minimum)

    These parameters provide ~100ms derivation time on modern hardware
    while maintaining strong resistance to offline password guessing.

    Args:
        password: The password string.
        salt: The salt bytes (must be 16 bytes).

    Returns:
        The derived key bytes (32 bytes).
    """
    key = hash_secret_raw(
        secret=password.encode(),
        salt=salt,
        time_cost=4,
        memory_cost=128 * 1024,  # 128 MiB
        parallelism=2,
        hash_len=32,
        type=Type.ID,
    )
    return key



"""CLI main entry point for Desktop 2FA."""

import os
import sys

import typer
from rich import print as rprint
from rich.text import Text

from desktop_2fa import __version__

from . import commands, helpers


def is_interactive() -> bool:
    """Check if we're running in interactive mode.

    For tests, this can be overridden with DESKTOP_2FA_FORCE_INTERACTIVE=1
    environment variable. For real usage, it uses TTY detection.
    """
    if os.getenv("DESKTOP_2FA_FORCE_INTERACTIVE") == "1":
        return True
    return sys.stdin.isatty() and sys.stdout.isatty()


app = typer.Typer(help="Desktop‑2FA — secure offline TOTP authenticator")


@app.callback(invoke_without_command=True)
def main(
    ctx: typer.Context,
    version: bool = typer.Option(False, "--version", help="Show version and exit"),
    password: str = typer.Option(
        None, "--password", help="Password for vault encryption/decryption"
    ),
    password_file: str = typer.Option(
        None,
        "--password-file",
        help="File containing password for vault encryption/decryption",
    ),
    allow_weak_passwords: bool = typer.Option(
        False, "--allow-weak-passwords", help="Allow weak passwords"
    ),
) -> None:
    """
    Global CLI callback — initializes context and handles --version and no-args case.

    Contract (from tests):
    - `desktop-2fa --version` → prints version, exit 0
    - `desktop-2fa` (no command) → prints version, exit 0
    - any command → ctx.obj must be initialized and no early exit
    """

    # ZAWSZE inicjalizujemy ctx.obj – niezależnie od komendy / opcji
    ctx.obj = {
        "password": password,
        "password_file": password_file,
        "interactive": is_interactive(),
        "allow_weak_passwords": allow_weak_passwords,
    }

    # Jeśli użytkownik podał --version LUB nie podał żadnej komendy,
    # zachowujemy się jak "print version and exit".
    # invoke_without_command=True gwarantuje, że callback jest wywołany także bez komendy.
    if version or ctx.invoked_subcommand is None:
        print(f"Desktop-2FA v{__version__}")
        raise typer.Exit()


@app.command("list")
def list_cmd(ctx: typer.Context) -> None:
    commands.list_entries(ctx)


@app.command("add")
def add_cmd(
    ctx: typer.Context,
    name: str = typer.Argument(None, help="Unique name identifier"),
    issuer: str = typer.Argument(None, help="Issuer name or otpauth:// URL"),
    secret: str = typer.Argument(None, help="TOTP secret (Base32)"),
) -> None:
    # Interactive mode: prompt for missing arguments
    interactive = ctx.obj.get("interactive", False)
    if interactive and (name is None or issuer is None or secret is None):
        if name is None:
            rprint(Text("Name (unique identifier):", style="cyan"))
            name = typer.prompt("")

        if issuer is None:
            rprint(Text("Issuer:", style="cyan"))
            issuer = typer.prompt("")

        if secret is None:
            rprint(Text("Secret:", style="cyan"))
            secret = typer.prompt("")

        commands.add_entry_interactive(name, issuer, secret, ctx)
        return

    if name is None or issuer is None or secret is None:
        helpers.print_error("Missing arguments: NAME, ISSUER and SECRET are required")
        rprint("Usage: d2fa add NAME ISSUER SECRET")
        rprint("Example: d2fa add GitHub GitHub ABCDEFGHIJKL1234")
        raise typer.Exit(1)

    commands.add_entry(name, issuer, secret, ctx)


@app.command("code")
def code_cmd(ctx: typer.Context, name: str) -> None:
    commands.generate_code(name, ctx)


@app.command("remove")
def remove_cmd(ctx: typer.Context, name: str) -> None:
    commands.remove_entry(name, ctx)


@app.command("rename")
def rename_cmd(ctx: typer.Context, old: str, new: str) -> None:
    commands.rename_entry(old, new, ctx)


@app.command("export")
def export_cmd(ctx: typer.Context, path: str) -> None:
    commands.export_vault(path, ctx)


@app.command("import")
def import_cmd(
    ctx: typer.Context,
    source: str,
    force: bool = typer.Option(False, "--force", help="Overwrite existing vault"),
) -> None:
    commands.import_vault(source, force, ctx)


@app.command("backup")
def backup_cmd(ctx: typer.Context) -> None:
    commands.backup_vault(ctx)


@app.command("init-vault")
def init_vault_cmd(
    ctx: typer.Context,
    force: bool = typer.Option(False, "--force", help="Overwrite existing vault"),
) -> None:
    """Initialize a new encrypted vault."""
    commands.init_vault(force, ctx)



import pathlib

import pytest

from desktop_2fa.cli.importers import (
    import_from_format,
    parse_1password_csv,
    parse_aegis_json,
    parse_bitwarden_csv,
    parse_freeotp_xml,
    parse_otpauth_uri,
)


def test_parse_aegis_json() -> None:
    """Test parsing Aegis JSON format."""
    content = """{
        "entries": [
            {
                "type": "totp",
                "issuer": "GitHub",
                "name": "user@github.com",
                "info": {
                    "secret": "JBSWY3DPEHPK3PXP",
                    "digits": 6,
                    "period": 30,
                    "algorithm": "SHA1"
                }
            },
            {
                "type": "hotp",
                "issuer": "Test",
                "name": "test",
                "info": {
                    "secret": "SECRET2"
                }
            },
            {
                "type": "totp",
                "issuer": "Empty",
                "name": "empty",
                "info": {
                    "secret": ""
                }
            }
        ]
    }"""

    entries = parse_aegis_json(content)
    assert len(entries) == 1
    entry = entries[0]
    assert entry["issuer"] == "GitHub"
    assert entry["account_name"] == "user@github.com"
    assert entry["secret"] == "JBSWY3DPEHPK3PXP"
    assert entry["digits"] == 6
    assert entry["period"] == 30
    assert entry["algorithm"] == "SHA1"


def test_parse_bitwarden_csv() -> None:
    """Test parsing Bitwarden CSV format."""
    content = """name,totp
GitHub,user@github.com,JBSWY3DPEHPK3PXP
Google,test@gmail.com,JBSWY3DPEHPK3PXS
Simple,JBSWY3DPEHPK3PXT
Empty,"""

    entries = parse_bitwarden_csv(content)
    assert len(entries) == 3

    # First entry with colon
    assert entries[0]["issuer"] == "GitHub"
    assert entries[0]["account_name"] == "user@github.com"
    assert entries[0]["secret"] == "JBSWY3DPEHPK3PXP"

    # Second entry with colon
    assert entries[1]["issuer"] == "Google"
    assert entries[1]["account_name"] == "test@gmail.com"
    assert entries[1]["secret"] == "JBSWY3DPEHPK3PXS"

    # Third entry without colon
    assert entries[2]["issuer"] == "Simple"
    assert entries[2]["account_name"] == "Simple"
    assert entries[2]["secret"] == "JBSWY3DPEHPK3PXT"


def test_parse_1password_csv() -> None:
    """Test parsing 1Password CSV format."""
    content = """title,otp
GitHub,JBSWY3DPEHPK3PXP
Google:test@gmail.com,JBSWY3DPEHPK3PXS
Empty,"""

    entries = parse_1password_csv(content)
    assert len(entries) == 2

    assert entries[0]["issuer"] == "GitHub"
    assert entries[0]["account_name"] == "GitHub"
    assert entries[0]["secret"] == "JBSWY3DPEHPK3PXP"

    assert entries[1]["issuer"] == "Google:test@gmail.com"
    assert entries[1]["account_name"] == "Google:test@gmail.com"
    assert entries[1]["secret"] == "JBSWY3DPEHPK3PXS"


def test_parse_otpauth_uri() -> None:
    """Test parsing otpauth URI."""
    uri = "otpauth://totp/GitHub:user@github.com?secret=JBSWY3DPEHPK3PXP&digits=6&period=30&algorithm=SHA1"

    entries = parse_otpauth_uri(uri)
    assert len(entries) == 1
    entry = entries[0]
    assert entry["issuer"] == "GitHub"
    assert entry["account_name"] == "user@github.com"
    assert entry["secret"] == "JBSWY3DPEHPK3PXP"
    assert entry["digits"] == 6
    assert entry["period"] == 30
    assert entry["algorithm"] == "SHA1"


def test_parse_otpauth_uri_no_issuer() -> None:
    """Test parsing otpauth URI without issuer."""
    uri = "otpauth://totp/user@github.com?secret=JBSWY3DPEHPK3PXP"

    entries = parse_otpauth_uri(uri)
    assert len(entries) == 1
    entry = entries[0]
    assert entry["issuer"] == ""
    assert entry["account_name"] == "user@github.com"
    assert entry["secret"] == "JBSWY3DPEHPK3PXP"


def test_parse_otpauth_uri_invalid() -> None:
    """Test parsing invalid otpauth URI."""
    with pytest.raises(ValueError, match="Invalid otpauth URI"):
        parse_otpauth_uri("https://example.com")

    with pytest.raises(ValueError, match="Not a TOTP otpauth URI"):
        parse_otpauth_uri("otpauth://hotp/GitHub:user?secret=SECRET")

    with pytest.raises(ValueError, match="No secret in otpauth URI"):
        parse_otpauth_uri("otpauth://totp/GitHub:user")


def test_parse_freeotp_xml() -> None:
    """Test parsing FreeOTP XML format."""
    content = """<?xml version="1.0" encoding="utf-8"?>
<tokens>
    <token>
        <issuer>GitHub</issuer>
        <label>user@github.com</label>
        <secret>JBSWY3DPEHPK3PXP</secret>
    </token>
    <token>
        <issuer></issuer>
        <label>Google:test@gmail.com</label>
        <secret>JBSWY3DPEHPK3PXS</secret>
    </token>
    <token>
        <issuer>Simple</issuer>
        <label>simple</label>
        <secret></secret>
    </token>
</tokens>"""

    entries = parse_freeotp_xml(content)
    assert len(entries) == 2

    assert entries[0]["issuer"] == "GitHub"
    assert entries[0]["account_name"] == "user@github.com"
    assert entries[0]["secret"] == "JBSWY3DPEHPK3PXP"

    assert entries[1]["issuer"] == "Google"
    assert entries[1]["account_name"] == "test@gmail.com"
    assert entries[1]["secret"] == "JBSWY3DPEHPK3PXS"


def test_import_from_format_aegis(tmp_path: pathlib.Path) -> None:
    """Test import_from_format with Aegis JSON."""
    aegis_file = tmp_path / "aegis.json"
    aegis_file.write_text(
        '{"entries": [{"type": "totp", "issuer": "Test", "name": "test", "info": {"secret": "SECRET"}}]}'
    )

    entries = import_from_format("aegis", str(aegis_file))
    assert len(entries) == 1
    assert entries[0]["issuer"] == "Test"


def test_import_from_format_bitwarden(tmp_path: pathlib.Path) -> None:
    """Test import_from_format with Bitwarden CSV."""
    csv_file = tmp_path / "bitwarden.csv"
    csv_file.write_text("name,totp\nTest,SECRET")

    entries = import_from_format("bitwarden", str(csv_file))
    assert len(entries) == 1
    assert entries[0]["issuer"] == "Test"


def test_import_from_format_1password(tmp_path: pathlib.Path) -> None:
    """Test import_from_format with 1Password CSV."""
    csv_file = tmp_path / "1password.csv"
    csv_file.write_text("title,otp\nTest,SECRET")

    entries = import_from_format("1password", str(csv_file))
    assert len(entries) == 1
    assert entries[0]["issuer"] == "Test"


def test_import_from_format_otpauth() -> None:
    """Test import_from_format with otpauth URI."""
    uri = "otpauth://totp/Test:test?secret=SECRET"
    entries = import_from_format("otpauth", uri)
    assert len(entries) == 1
    assert entries[0]["issuer"] == "Test"


def test_import_from_format_freeotp(tmp_path: pathlib.Path) -> None:
    """Test import_from_format with FreeOTP XML."""
    xml_file = tmp_path / "freeotp.xml"
    xml_file.write_text(
        """<?xml version="1.0"?>
<tokens>
    <token>
        <issuer>Test</issuer>
        <label>test</label>
        <secret>SECRET</secret>
    </token>
</tokens>"""
    )

    entries = import_from_format("freeotp", str(xml_file))
    assert len(entries) == 1
    assert entries[0]["issuer"] == "Test"


def test_import_from_format_unsupported() -> None:
    """Test import_from_format with unsupported format."""
    with pytest.raises(ValueError, match="Unsupported format"):
        import_from_format("unsupported", "dummy")



import os

import pytest

from desktop_2fa.crypto.aesgcm import decrypt, encrypt
from desktop_2fa.crypto.argon2 import derive_key


def test_encrypt_decrypt() -> None:
    salt = os.urandom(16)
    key = derive_key("password", salt)
    data = b"hello"
    enc = encrypt(key, data)
    dec = decrypt(key, enc)
    assert dec == data


def test_decrypt_corrupted_ciphertext() -> None:
    """Test decryption with corrupted ciphertext that passes length validation."""
    salt = os.urandom(16)
    key = derive_key("password", salt)
    data = b"hello"
    enc = encrypt(key, data)

    # Corrupt the ciphertext by flipping a bit
    corrupted = bytearray(enc)
    corrupted[10] ^= 0xFF

    with pytest.raises(ValueError):
        decrypt(key, bytes(corrupted))


def test_encrypt_decrypt_empty_data() -> None:
    """Test encryption and decryption with empty data."""
    salt = os.urandom(16)
    key = derive_key("password", salt)
    data = b""
    enc = encrypt(key, data)
    dec = decrypt(key, enc)
    assert dec == data


def test_encrypt_decrypt_large_data() -> None:
    """Test encryption and decryption with large data."""
    salt = os.urandom(16)
    key = derive_key("password", salt)
    data = b"x" * 10000  # 10KB of data
    enc = encrypt(key, data)
    dec = decrypt(key, enc)
    assert dec == data


def test_decrypt_too_short_blob() -> None:
    """Test decryption with blob too short for nonce."""
    salt = os.urandom(16)
    key = derive_key("password", salt)

    with pytest.raises(ValueError, match="Encrypted blob too short"):
        decrypt(key, b"short")



import re
import shutil
from pathlib import Path
from typing import Any

import pytest
from typer.testing import CliRunner

from desktop_2fa.cli.helpers import load_vault, save_vault
from desktop_2fa.cli.main import app

TEST_PASSWORD = "jawislajawisla"
runner = CliRunner()


@pytest.fixture
def fake_vault_env_cli(tmp_path: Path, monkeypatch: Any) -> Path:
    fake_vault = tmp_path / "vault"

    monkeypatch.setattr(
        "desktop_2fa.cli.helpers.get_vault_path",
        lambda: str(fake_vault),
    )

    if fake_vault.parent.exists():
        shutil.rmtree(fake_vault.parent)
    fake_vault.parent.mkdir(parents=True, exist_ok=True)

    return fake_vault


def test_cli_list_empty(fake_vault_env_cli: Path) -> None:
    result = runner.invoke(app, ["--password", TEST_PASSWORD, "list"])
    assert result.exit_code == 0
    # helpers.list_entries prints nothing when empty
    assert result.output.strip() == ""


def test_cli_add_and_list(fake_vault_env_cli: Path) -> None:
    result = runner.invoke(
        app,
        ["--password", TEST_PASSWORD, "add", "GitHub", "GitHub", "JBSWY3DPEHPK3PXP"],
    )
    assert result.exit_code == 0
    assert "Entry added: GitHub" in result.output

    result = runner.invoke(app, ["--password", TEST_PASSWORD, "list"])
    assert result.exit_code == 0
    assert "- GitHub (GitHub)" in result.output


def test_cli_code(fake_vault_env_cli: Path) -> None:
    runner.invoke(
        app,
        ["--password", TEST_PASSWORD, "add", "GitHub", "GitHub", "JBSWY3DPEHPK3PXP"],
    )
    result = runner.invoke(app, ["--password", TEST_PASSWORD, "code", "GitHub"])
    assert result.exit_code == 0
    code = result.output.strip()
    assert len(code) == 6
    assert code.isdigit()


def test_cli_remove(fake_vault_env_cli: Path) -> None:
    runner.invoke(
        app,
        ["--password", TEST_PASSWORD, "add", "GitHub", "GitHub", "JBSWY3DPEHPK3PXP"],
    )
    result = runner.invoke(app, ["--password", TEST_PASSWORD, "remove", "GitHub"])
    assert result.exit_code == 0
    assert "Removed entry: GitHub" in result.output


def test_cli_rename(fake_vault_env_cli: Path) -> None:
    runner.invoke(
        app,
        ["--password", TEST_PASSWORD, "add", "GitHub", "GitHub", "JBSWY3DPEHPK3PXP"],
    )
    result = runner.invoke(
        app,
        ["--password", TEST_PASSWORD, "rename", "GitHub", "NewGitHub"],
    )
    assert result.exit_code == 0
    assert "Renamed 'GitHub' → 'NewGitHub'" in result.output


def test_cli_export(fake_vault_env_cli: Path, tmp_path: Path) -> None:
    runner.invoke(
        app,
        ["--password", TEST_PASSWORD, "add", "GitHub", "GitHub", "JBSWY3DPEHPK3PXP"],
    )
    export_path = tmp_path / "export.json"
    result = runner.invoke(
        app,
        ["--password", TEST_PASSWORD, "export", str(export_path)],
    )
    assert result.exit_code == 0
    assert "Exported vault to:" in result.output
    assert export_path.exists()


def test_cli_import(fake_vault_env_cli: Path, tmp_path: Path) -> None:
    runner.invoke(
        app,
        ["--password", TEST_PASSWORD, "add", "GitHub", "GitHub", "JBSWY3DPEHPK3PXP"],
    )
    export_path = tmp_path / "export.json"
    runner.invoke(
        app,
        ["--password", TEST_PASSWORD, "export", str(export_path)],
    )

    # Use the fake vault path instead of the default vault path
    vault = load_vault(fake_vault_env_cli, TEST_PASSWORD)
    vault.entries.clear()
    save_vault(fake_vault_env_cli, vault, TEST_PASSWORD)

    result = runner.invoke(
        app,
        ["--password", TEST_PASSWORD, "import", "--force", str(export_path)],
    )
    assert result.exit_code == 0
    assert "Vault imported from" in result.output


def test_cli_backup(fake_vault_env_cli: Path) -> None:
    runner.invoke(
        app,
        ["--password", TEST_PASSWORD, "add", "GitHub", "GitHub", "JBSWY3DPEHPK3PXP"],
    )
    result = runner.invoke(app, ["--password", TEST_PASSWORD, "backup"])
    assert result.exit_code == 0
    assert "Backup created:" in result.output


def test_cli_version(fake_vault_env_cli: Path) -> None:
    result = runner.invoke(app, ["--version"])
    assert result.exit_code == 0
    assert "Desktop-2FA v" in result.output


def test_cli_version_no_args(fake_vault_env_cli: Path) -> None:
    result = runner.invoke(app, [])
    assert result.exit_code == 0
    assert "Desktop-2FA v" in result.output


def test_cli_add_missing_arguments(fake_vault_env_cli: Path) -> None:
    """Test add command with missing arguments."""
    result = runner.invoke(app, ["--password", TEST_PASSWORD, "add"])
    assert result.exit_code == 1
    assert "Missing arguments: NAME, ISSUER and SECRET are required" in result.output
    assert "Usage: d2fa add NAME ISSUER SECRET" in result.output
    assert "Example: d2fa add GitHub GitHub ABCDEFGHIJKL1234" in result.output


def test_cli_add_missing_issuer(fake_vault_env_cli: Path) -> None:
    """Test add command with missing issuer."""
    result = runner.invoke(
        app, ["--password", TEST_PASSWORD, "add", "Test", "", "JBSWY3DPEHPK3PXP"]
    )
    assert result.exit_code == 0  # Empty issuer is accepted
    assert "Entry added:" in result.output


def test_cli_add_missing_secret(fake_vault_env_cli: Path) -> None:
    """Test add command with missing secret."""
    result = runner.invoke(app, ["--password", TEST_PASSWORD, "add", "GitHub", "GitHub", ""])
    assert result.exit_code == 0  # Empty secret is accepted
    assert "Entry added:" in result.output


def test_cli_init_vault_new(fake_vault_env_cli: Path) -> None:
    """Test initializing new vault."""
    result = runner.invoke(app, ["--password", TEST_PASSWORD, "init-vault"])
    assert result.exit_code == 0
    assert "Vault created." in result.output
    assert fake_vault_env_cli.exists()


def test_cli_init_vault_existing_no_force(fake_vault_env_cli: Path) -> None:
    """Test initializing vault when it already exists without force."""
    from desktop_2fa.vault import Vault

    vault = Vault()
    vault.save(fake_vault_env_cli, TEST_PASSWORD)

    result = runner.invoke(app, ["--password", TEST_PASSWORD, "init-vault"])
    assert result.exit_code == 0
    assert "Vault already exists." in result.output
    assert "Use --force to overwrite." in result.output


def test_cli_init_vault_existing_with_force(fake_vault_env_cli: Path) -> None:
    """Test initializing vault when it already exists with force."""
    from desktop_2fa.vault import Vault

    vault = Vault()
    vault.save(fake_vault_env_cli, TEST_PASSWORD)

    result = runner.invoke(app, ["--password", TEST_PASSWORD, "init-vault", "--force"])
    assert result.exit_code == 0
    assert "Vault created." in result.output


def test_cli_is_interactive_tty_detection(monkeypatch: Any) -> None:
    """Test is_interactive function with TTY detection."""
    # Mock sys.stdin.isatty and sys.stdout.isatty to return True
    monkeypatch.setattr("sys.stdin.isatty", lambda: True)
    monkeypatch.setattr("sys.stdout.isatty", lambda: True)

    from desktop_2fa.cli.main import is_interactive

    assert is_interactive() is True


def test_cli_is_interactive_non_tty(monkeypatch: Any) -> None:
    """Test is_interactive function with non-TTY."""
    # Mock sys.stdin.isatty and sys.stdout.isatty to return False
    monkeypatch.setattr("sys.stdin.isatty", lambda: False)
    monkeypatch.setattr("sys.stdout.isatty", lambda: False)

    from desktop_2fa.cli.main import is_interactive

    assert is_interactive() is False


def test_cli_is_interactive_force_environment(monkeypatch: Any) -> None:
    """Test is_interactive function with force environment variable."""
    monkeypatch.setenv("DESKTOP_2FA_FORCE_INTERACTIVE", "1")

    from desktop_2fa.cli.main import is_interactive

    assert is_interactive() is True


def test_cli_interactive_empty_input(
    fake_vault_env_cli: Path, monkeypatch: Any
) -> None:
    """Test CLI interactive mode with empty input."""

    # Mock typer.prompt to return empty string
    def mock_prompt(*args: Any, **kwargs: Any) -> str:
        return ""

    monkeypatch.setattr("typer.prompt", mock_prompt)

    # This should handle empty input gracefully
    result = runner.invoke(
        app, ["--password", TEST_PASSWORD, "add", "GitHub", "GitHub", "JBSWY3DPEHPK3PXP"]
    )
    assert result.exit_code == 0  # Should not crash


def test_cli_interactive_whitespace_input(
    fake_vault_env_cli: Path, monkeypatch: Any
) -> None:
    """Test CLI interactive mode with whitespace-only input."""

    # Mock typer.prompt to return whitespace
    def mock_prompt(*args: Any, **kwargs: Any) -> str:
        return "   "

    monkeypatch.setattr("typer.prompt", mock_prompt)

    # This should handle whitespace input gracefully
    result = runner.invoke(
        app, ["--password", TEST_PASSWORD, "add", "GitHub", "GitHub", "JBSWY3DPEHPK3PXP"]
    )
    assert result.exit_code == 0  # Should not crash


def test_cli_add_interactive_prompts(
    fake_vault_env_cli: Path, monkeypatch: Any
) -> None:
    """Test add command interactive prompts."""
    # Force interactive mode
    monkeypatch.setenv("DESKTOP_2FA_FORCE_INTERACTIVE", "1")

    # Mock typer.prompt to return values for name, issuer and secret
    responses = ["TestName", "TestIssuer", "JBSWY3DPEHPK3PXP"]
    monkeypatch.setattr("typer.prompt", lambda text, hide_input=False: responses.pop(0))

    result = runner.invoke(app, ["--password", TEST_PASSWORD, "add"])
    assert result.exit_code == 0
    assert "Entry added: TestName" in result.output


def test_no_rich_markup_in_prompts() -> None:
    """Test that no prompt labels contain Rich markup."""
    # Patterns to search for
    forbidden_patterns = [
        r"\[cyan\]",
        r"\[red\]",
        r"\[green\]",
        r"\[yellow\]",
        r"\[magenta\]",
        r"\[blue\]",
        r"\[white\]",
        r"\[black\]",
    ]

    # Files to check
    src_dir = Path(__file__).parent.parent / "src"
    py_files = list(src_dir.rglob("*.py"))

    violations = []
    for file_path in py_files:
        try:
            content = file_path.read_text(encoding="utf-8")
            for pattern in forbidden_patterns:
                matches = re.findall(pattern, content)
                if matches:
                    violations.append(
                        f"{file_path}: {pattern} found {len(matches)} times"
                    )
        except Exception as e:
            violations.append(f"Error reading {file_path}: {e}")

    if violations:
        pytest.fail("Found Rich markup in prompts:\n" + "\n".join(violations))



"""Vault implementation for storing and managing TOTP entries."""

import os
import re
from pathlib import Path
from typing import Optional

from pydantic import ValidationError

from ..crypto.aesgcm import decrypt, encrypt
from ..crypto.argon2 import derive_key
from .models import TotpEntry, VaultData

# Vault file format constants
VAULT_MAGIC = b"D2FA"
VAULT_VERSION = b"\x01"
HEADER_LEN = len(VAULT_MAGIC) + len(VAULT_VERSION)


class VaultError(Exception):
    """Base exception for vault-related errors."""

    pass


class InvalidPassword(VaultError):
    """Raised when the provided password is incorrect."""

    pass


class CorruptedVault(VaultError):
    """Raised when the vault file is corrupted or contains invalid data."""

    pass


class UnsupportedFormat(VaultError):
    """Raised when the vault file format is unsupported or invalid."""

    pass


class VaultIOError(VaultError):
    """Raised for filesystem or IO-related errors."""

    pass


class Vault:
    """Vault using Pydantic models for validation and structure."""

    def __init__(self, data: Optional[VaultData] = None):
        """Initialize the vault with optional data.

        Args:
            data: The vault data to initialize with.
        """
        self.data = data or VaultData()

    @property
    def entries(self) -> list[TotpEntry]:
        """Get the list of TOTP entries.

        Returns:
            The list of TOTP entries.
        """
        return self.data.entries

    def add_entry(
        self,
        name: str,
        issuer: str,
        secret: str,
    ) -> None:
        """Add a new TOTP entry to the vault.

        Args:
            name: The unique account name.
            issuer: The issuer name.
            secret: The base32-encoded secret.
        """
        # Check for duplicate names
        for entry in self.entries:
            if entry.account_name == name:
                raise ValueError(f"An entry with name \"{name}\" already exists. Names must be unique.")

        entry = TotpEntry(
            account_name=name,
            issuer=issuer,
            secret=secret,
        )
        self.data.entries.append(entry)

    def rename_entry(self, old: str, new: str) -> None:
        """Rename an entry (placeholder for future implementation)."""
        raise NotImplementedError("rename will be implemented in a future version")

    def get_entry(self, issuer: str) -> TotpEntry:
        """Get a TOTP entry by issuer or account name.

        Args:
            issuer: The issuer or account name to search for.

        Returns:
            The matching TOTP entry.

        Raises:
            ValueError: If no entry is found.
        """
        for entry in self.entries:
            if entry.issuer == issuer or entry.account_name == issuer:
                return entry
        raise ValueError(f"Entry '{issuer}' not found")

    def remove_entry(self, issuer: str) -> None:
        """Remove a TOTP entry by issuer or account name.

        Args:
            issuer: The issuer or account name of the entry to remove.
        """
        entry = self.get_entry(issuer)
        self.data.entries.remove(entry)

    @classmethod
    def load(cls, path: str | Path, password: Optional[str] = None) -> "Vault":
        """Load a vault from a file.

        Args:
            path: The file path to load from.
            password: The password to decrypt the vault. If None, a default
                internal password is used (for "no-password" vaults).

        Returns:
            The loaded Vault instance.

        Raises:
            VaultIOError: If the vault file cannot be read due to IO errors.
            UnsupportedFormat: If the vault file format is invalid.
            InvalidPassword: If the password is incorrect.
            CorruptedVault: If the vault data is corrupted.
        """
        try:
            with open(path, "rb") as f:
                blob = f.read()
        except OSError as e:
            raise VaultIOError(f"Failed to read vault file: {e}") from e

        if len(blob) < HEADER_LEN + 16:
            raise UnsupportedFormat("Vault file is too short or invalid format")

        # Check magic header and version
        if blob[:4] != VAULT_MAGIC:
            raise UnsupportedFormat("Invalid vault file format: incorrect magic header")
        if blob[4:5] != VAULT_VERSION:
            raise UnsupportedFormat("Unsupported vault file version")

        # First 5 bytes: header, next 16: salt, rest: AES-GCM blob (nonce + ciphertext + tag)
        salt, encrypted = blob[HEADER_LEN : HEADER_LEN + 16], blob[HEADER_LEN + 16 :]

        if len(encrypted) == 0:
            raise UnsupportedFormat("Vault file is invalid: empty encrypted blob")

        # For "no-password" vaults we consistently use an empty password string.
        if password is None:
            password = ""

        key = derive_key(password, salt)
        try:
            raw_json = decrypt(key, encrypted)
        except ValueError as e:
            raise InvalidPassword("Invalid password or corrupted vault") from e

        # decrypt() zwraca bytes; Pydantic v2 akceptuje bytes jako JSON input.
        try:
            data = VaultData.model_validate_json(raw_json)
        except ValidationError as e:
            # If Pydantic validation fails, it's corrupted data (not a password issue)
            # because the decryption succeeded but the data structure is wrong
            raise CorruptedVault("Vault contains invalid data") from e

        vault = cls(data)

        # Check for duplicate account names and warn
        names = [entry.account_name for entry in vault.entries if entry.account_name]
        duplicates = [name for name in set(names) if names.count(name) > 1]
        if duplicates:
            print(f"Warning: Your vault contains multiple entries with the same name: {', '.join(chr(34) + name + chr(34) for name in duplicates)}.")
            print("This was allowed in older versions. You can resolve this by renaming entries using the rename command.")

        return vault

    def save(self, path: str | Path, password: Optional[str] = None) -> None:
        """Save the vault to a file.

        Args:
            path: The file path to save to.
            password: The password to encrypt the vault. If None, a default
                internal password is used (for "no-password" vaults).

        Raises:
            VaultIOError: If saving fails due to IO errors.
        """
        path = Path(path)
        path.parent.mkdir(parents=True, exist_ok=True)

        temp_path = path.with_suffix(".tmp")

        try:
            header = VAULT_MAGIC + VAULT_VERSION
            salt = os.urandom(16)

            # For "no-password" vaults we consistently use an empty password string.
            if password is None:
                password = ""

            key = derive_key(password, salt)

            raw_json = self.data.model_dump_json().encode("utf-8")
            encrypted = encrypt(key, raw_json)

            fd = os.open(str(temp_path), os.O_WRONLY | os.O_CREAT | os.O_TRUNC, 0o600)
            try:
                with os.fdopen(fd, "wb") as f:
                    f.write(header + salt + encrypted)
                    f.flush()
                    os.fsync(fd)
            except:
                os.close(fd)
                raise

            os.replace(temp_path, path)
        except OSError as e:
            if temp_path.exists():
                try:
                    temp_path.unlink()
                except OSError:
                    pass
            # Extract just the error message without errno prefix
            error_msg = str(e)
            if error_msg.startswith("[Errno"):
                # Extract the message part after the errno
                match = re.search(r"\] (.+)$", error_msg)
                if match:
                    error_msg = match.group(1)
            raise VaultIOError(f"Failed to save vault: {error_msg}") from e



import pytest

from desktop_2fa.totp.generator import generate

SECRET = "JBSWY3DPEHPK3PXP"  # poprawny base32


def test_totp_sha1_deterministic() -> None:
    code1 = generate(SECRET, timestamp=0, digits=6, period=30, algorithm="SHA1")
    code2 = generate(SECRET, timestamp=0, digits=6, period=30, algorithm="SHA1")

    assert code1 == code2
    assert len(code1) == 6
    assert code1.isdigit()


def test_totp_different_time_produces_different_code() -> None:
    code1 = generate(SECRET, timestamp=0, digits=6, period=30, algorithm="SHA1")
    code2 = generate(SECRET, timestamp=60, digits=6, period=30, algorithm="SHA1")

    assert code1 != code2


def test_totp_sha256_and_sha512() -> None:
    code256 = generate(SECRET, timestamp=0, digits=6, period=30, algorithm="SHA256")
    code512 = generate(SECRET, timestamp=0, digits=6, period=30, algorithm="SHA512")

    assert len(code256) == 6
    assert len(code512) == 6
    assert code256.isdigit()
    assert code512.isdigit()
    # raczej różne od SHA1 dla tego samego czasu
    code1 = generate(SECRET, timestamp=0, digits=6, period=30, algorithm="SHA1")
    assert code256 != code1 or code512 != code1


def test_totp_different_digits() -> None:
    code6 = generate(SECRET, timestamp=0, digits=6, period=30, algorithm="SHA1")
    code8 = generate(SECRET, timestamp=0, digits=8, period=30, algorithm="SHA1")

    assert len(code6) == 6
    assert len(code8) == 8
    assert code6.isdigit()
    assert code8.isdigit()


def test_totp_unsupported_algorithm() -> None:
    with pytest.raises(ValueError):
        generate(SECRET, timestamp=0, digits=6, period=30, algorithm="MD5")



"""CLI helper functions for Desktop 2FA."""

import base64
import math
import os
import tomllib
import urllib.parse
from pathlib import Path
from typing import TYPE_CHECKING, Any

import typer
from rich import print as rprint
from rich.text import Text

from desktop_2fa.vault import Vault

if TYPE_CHECKING:
    from desktop_2fa.vault.models import TotpEntry


def list_entries(path: Path, password: str) -> None:
    """List all entries in the vault."""
    vault = Vault.load(path, password)
    for entry in vault.entries:
        print(f"- {entry.account_name} ({entry.issuer})")


def add_entry(
    path: Path, issuer: str, account: str, secret: str, password: str
) -> None:
    """Add a new entry to the vault."""
    vault = Vault.load(path, password)
    vault.add_entry(name=account, issuer=issuer, secret=secret)
    vault.save(path, password)
    print(f"Added entry: {account}")


def generate_code(path: Path, name: str, password: str) -> None:
    """Generate and print the TOTP code for the given issuer."""
    vault = Vault.load(path, password)
    entry = vault.get_entry(name)

    from desktop_2fa.totp.generator import generate

    code = generate(
        secret=entry.secret,
        digits=entry.digits,
        period=entry.period,
        algorithm=entry.algorithm,
    )

    print(code)


def remove_entry(path: Path, name: str, password: str) -> None:
    """Remove an entry from the vault."""
    vault = Vault.load(path, password)
    vault.remove_entry(name)
    vault.save(path, password)
    print(f"Removed entry: {name}")


def rename_entry(path: Path, old: str, new: str, password: str) -> None:
    """Rename an entry."""
    vault = Vault.load(path, password)
    entry = vault.get_entry(old)
    entry.account_name = new
    entry.issuer = new

    vault.save(path, password)
    print(f"Renamed '{old}' → '{new}'")


def export_vault(path: Path, export_path: Path, password: str) -> None:
    """Export the vault file."""
    vault = Vault.load(path, password)
    vault.save(export_path, password)
    print(f"Exported vault to: {export_path}")


def import_vault(path: Path, import_path: Path, password: str) -> None:
    """Import the vault file."""
    vault = Vault.load(import_path, password=password)
    vault.save(path, password)
    print("Vault imported from")


def backup_vault(path: Path, backup_path: Path, password: str) -> None:
    """Create a backup of the vault file."""
    vault = Vault.load(path, password)
    vault.save(backup_path, password)
    print("Backup created:")


def get_vault_path() -> str:
    """Get the default path for the vault file."""
    return str(Path.home() / ".desktop-2fa" / "vault")


def load_vault(path: Path, password: str) -> Vault:
    """Load the vault from the specified path."""
    return Vault.load(path, password)


def save_vault(path: Path, vault: Vault, password: str) -> None:
    """Save the vault to the specified path."""
    vault.save(path, password)


def get_password_for_vault(ctx: typer.Context, new_vault: bool = False) -> str:
    """Get the password for vault operations."""
    password = ctx.obj.get("password")
    password_file = ctx.obj.get("password_file")
    interactive = ctx.obj.get("interactive")

    # Both flags provided
    if password and password_file:
        print("Error: Cannot specify both --password and --password-file")
        raise typer.Exit(1)

    # Direct password
    if password:
        return password  # type: ignore[no-any-return]

    # Password from file
    if password_file:
        try:
            with open(password_file, "r") as f:
                return f.read().strip()
        except FileNotFoundError:
            print(f"Error: Password file '{password_file}' not found")
            raise typer.Exit(1)
        except Exception as e:
            print(f"Error reading password file: {e}")
            raise typer.Exit(1)

    # No password provided
    if not interactive:
        print("Error: Password not provided and not running in interactive mode")
        raise typer.Exit(1)

    # Interactive mode → prompt
    if new_vault:
        rprint(Text("Enter new vault password:", style="cyan"))
        pwd = typer.prompt("", hide_input=True)
        rprint(Text("Confirm vault password:", style="cyan"))
        confirm = typer.prompt("", hide_input=True)
        if pwd != confirm:
            print_error("Passwords do not match. Please try again.")
            raise typer.Exit(1)
        # Check password strength
        if not _should_skip_password_checks(ctx):
            _enforce_password_strength(pwd)
    else:
        rprint(Text("Enter vault password:", style="cyan"))
        pwd = typer.prompt("", hide_input=True)
    return pwd  # type: ignore[no-any-return]


def _read_password_file(password_file: str) -> str:
    """Read password from file."""
    try:
        with open(password_file, "r") as f:
            return f.read().strip()
    except FileNotFoundError:
        print(f"Error: Password file '{password_file}' not found")
        raise typer.Exit(1)
    except Exception as e:
        print(f"Error reading password file: {e}")
        raise typer.Exit(1)


def load_config() -> dict[str, Any]:
    """Load configuration from ~/.config/d2fa/config.toml."""
    config_path = Path.home() / ".config" / "d2fa" / "config.toml"
    if not config_path.exists():
        return {}
    with open(config_path, "rb") as f:
        return tomllib.load(f)


def calculate_entropy(password: str) -> float:
    """Calculate the entropy of a password."""
    words = password.split()
    if len(words) >= 4:
        return 11 * len(words)
    # Determine character set size
    has_lower = any(c.islower() for c in password)
    has_upper = any(c.isupper() for c in password)
    has_digit = any(c.isdigit() for c in password)
    has_symbol = any(not c.isalnum() for c in password)
    N = 0
    if has_lower:
        N += 26
    if has_upper:
        N += 26
    if has_digit:
        N += 10
    if has_symbol:
        N += 32  # approximate
    if N == 0:
        N = 1
    return len(password) * math.log2(N)


def _should_skip_password_checks(ctx: typer.Context) -> bool:
    """Check if password strength checks should be skipped."""
    return (
        ctx.obj.get("allow_weak_passwords", False)
        or os.getenv("D2FA_ALLOW_WEAK_PASSWORDS") == "1"
        or os.getenv("PYTEST_CURRENT_TEST") is not None
    )


def _enforce_password_strength(password: str) -> None:
    """Enforce password strength requirements."""
    config = load_config()
    security = config.get("security", {})
    min_entropy = security.get("min_password_entropy", 60)
    reject_weak = security.get("reject_weak_passwords", False)

    entropy = calculate_entropy(password)
    if entropy < min_entropy:
        if reject_weak:
            print_error(
                f"Password too weak (entropy {entropy:.1f} < {min_entropy}). "
                "Please choose a stronger password."
            )
            raise typer.Exit(1)
        else:
            print_warning(
                f"Password is weak (entropy {entropy:.1f} < {min_entropy}). "
                "Consider using a stronger password."
            )
            if not typer.confirm("Continue with weak password?"):
                raise typer.Exit(1)


def get_password_from_cli(ctx: typer.Context) -> str:
    """Legacy alias for backward compatibility."""
    return get_password_for_vault(ctx, new_vault=False)


# Rich-based output helpers
def print_success(message: str) -> None:
    """Print a success message in green."""
    rprint(Text(message, style="green"))


def print_warning(message: str) -> None:
    """Print a warning message in yellow."""
    rprint(Text(message, style="yellow"))


def print_error(message: str) -> None:
    """Print an error message in red."""
    rprint(Text(message, style="red"))


def print_info(message: str) -> None:
    """Print an info message in white."""
    rprint(Text(message, style="white"))


def print_prompt(message: str) -> None:
    """Print a prompt message in cyan."""
    rprint(Text(message, style="cyan"))


def print_header(message: str) -> None:
    """Print a header message in bold white."""
    rprint(Text(message, style="bold white"))


def print_entries_table(entries: list["TotpEntry"]) -> None:
    """Print entries in a formatted table."""
    if not entries:
        print_info("No entries found.")
        return

    # For now, print simple format to keep tests passing
    for entry in entries:
        print(f"- {entry.account_name} ({entry.issuer})")


def validate_base32(secret: str) -> bool:
    """Validate if a string is valid Base32."""
    try:
        # Remove padding and spaces, convert to uppercase
        cleaned = secret.replace(" ", "").replace("=", "").upper()
        # Check if all characters are valid Base32
        if not all(c in "ABCDEFGHIJKLMNOPQRSTUVWXYZ234567" for c in cleaned):
            return False
        # Try to decode to verify it's valid
        base64.b32decode(cleaned + "=" * ((8 - len(cleaned) % 8) % 8))
        return True
    except Exception:
        return False


def parse_otpauth_url(url: str) -> dict[str, str]:
    """Parse an otpauth:// URL and extract issuer, label, and secret."""
    if not url.startswith("otpauth://"):
        raise ValueError("Invalid otpauth URL")

    parsed = urllib.parse.urlparse(url)
    if parsed.scheme != "otpauth" or parsed.netloc != "totp":
        raise ValueError("Only TOTP otpauth URLs are supported")

    # Parse the path: /Issuer:Label or /Issuer or /:Label
    path = parsed.path.lstrip("/")
    if ":" in path:
        issuer, label = path.split(":", 1)
    else:
        issuer = path
        label = path

    # Parse query parameters
    query = urllib.parse.parse_qs(parsed.query)
    secret = query.get("secret", [None])[0]
    if not secret:
        raise ValueError("Secret parameter is required")

    url_issuer = query.get("issuer", [None])[0]
    if url_issuer and not issuer:
        issuer = url_issuer

    return {
        "issuer": issuer or "Unknown",
        "label": label or issuer or "Unknown",
        "secret": secret,
    }





from pathlib import Path

import pytest

from desktop_2fa.vault.vault import Vault


def test_vault_roundtrip(tmp_path: Path) -> None:
    path = tmp_path / "vault.bin"

    vault = Vault()
    vault.add_entry("GitHub", "GitHub", "JBSWY3DPEHPK3PXP")
    vault.save(str(path))

    loaded = Vault.load(str(path))
    assert loaded.get_entry("GitHub").secret == "JBSWY3DPEHPK3PXP"


def test_vault_is_binary(tmp_path: Path) -> None:
    path = tmp_path / "vault.bin"

    vault = Vault()
    vault.add_entry("Test", "Test", "JBSWY3DPEHPK3PXP")
    vault.save(str(path))

    raw = path.read_bytes()
    assert raw != b""
    assert b"Test" not in raw
    assert b"JBSWY3DPEHPK3PXP" not in raw


def test_vault_backup_created(tmp_path: Path) -> None:
    path = tmp_path / "vault.bin"

    vault = Vault()
    vault.add_entry("A", "A", "JBSWY3DPEHPK3PXP")
    vault.save(str(path))

    vault.add_entry("B", "B", "JBSWY3DPEHPK3PXS")
    vault.save(str(path))

    # Note: New vault implementation does not create backups
    # assert backup.exists()
    # assert backup.read_bytes() != b""


def test_vault_tampering_detection(tmp_path: Path) -> None:
    path = tmp_path / "vault.bin"

    vault = Vault()
    vault.add_entry("X", "X", "JBSWY3DPEHPK3PXP")
    vault.save(str(path))

    raw = bytearray(path.read_bytes())
    raw[10] ^= 0xFF  # celowa korupcja
    path.write_bytes(raw)

    with pytest.raises(Exception):
        Vault.load(str(path))


def test_multiple_entries(tmp_path: Path) -> None:
    path = tmp_path / "vault.bin"

    vault = Vault()
    vault.add_entry("GitHub", "GitHub", "JBSWY3DPEHPK3PXP")
    vault.add_entry("Google", "Google", "JBSWY3DPEHPK3PXS")
    vault.add_entry("AWS", "AWS", "JBSWY3DPEHPK3PXT")
    vault.save(str(path))

    loaded = Vault.load(str(path))
    assert loaded.get_entry("GitHub").secret == "JBSWY3DPEHPK3PXP"
    assert loaded.get_entry("Google").secret == "JBSWY3DPEHPK3PXS"
    assert loaded.get_entry("AWS").secret == "JBSWY3DPEHPK3PXT"


def test_vault_encryption_with_password(tmp_path: Path) -> None:
    """Test vault encryption and decryption with user-supplied password."""
    path = tmp_path / "vault_encrypted.bin"
    password = "test_password_123"

    vault = Vault()
    vault.add_entry("GitHub", "GitHub", "JBSWY3DPEHPK3PXP")
    vault.add_entry("Google", "Google", "JBSWY3DPEHPK3PXS")
    vault.save(str(path), password=password)

    # Verify the file is encrypted (not plaintext)
    raw = path.read_bytes()
    assert raw != b""
    assert b"GitHub" not in raw
    assert b"JBSWY3DPEHPK3PXP" not in raw

    # Load with correct password
    loaded = Vault.load(str(path), password=password)
    assert loaded.get_entry("GitHub").secret == "JBSWY3DPEHPK3PXP"
    assert loaded.get_entry("Google").secret == "JBSWY3DPEHPK3PXS"


def test_vault_wrong_password(tmp_path: Path) -> None:
    """Test that vault cannot be loaded with wrong password."""
    path = tmp_path / "vault_wrong_pass.bin"
    correct_password = "correct_pass"
    wrong_password = "wrong_pass"

    vault = Vault()
    vault.add_entry("Test", "Test", "JBSWY3DPEHPK3PXP")
    vault.save(str(path), password=correct_password)

    # Should fail with wrong password
    with pytest.raises(Exception):
        Vault.load(str(path), password=wrong_password)


def test_vault_password_roundtrip(tmp_path: Path) -> None:
    """Test multiple save/load cycles with password."""
    path = tmp_path / "vault_roundtrip.bin"
    password = "roundtrip_pass"

    # Base32-sekrety
    secret1 = "JBSWY3DPEHPK3PXP"
    secret2 = "JBSWY3DPEHPK3PXS"

    vault1 = Vault()
    vault1.add_entry("Entry1", "Entry1", secret1)
    vault1.save(str(path), password=password)

    vault2 = Vault.load(str(path), password=password)
    vault2.add_entry("Entry2", "Entry2", secret2)
    vault2.save(str(path), password=password)

    vault3 = Vault.load(str(path), password=password)
    assert vault3.get_entry("Entry1").secret == secret1
    assert vault3.get_entry("Entry2").secret == secret2


def test_vault_no_password_vs_password(tmp_path: Path) -> None:
    """Test that password-protected vaults are different from non-protected ones."""
    path_no_pass = tmp_path / "vault_no_pass.bin"
    path_with_pass = tmp_path / "vault_with_pass.bin"
    password = "some_password"

    vault = Vault()
    vault.add_entry("Test", "Test", "JBSWY3DPEHPK3PXP")

    # Save without password
    vault.save(str(path_no_pass))

    # Save with password
    vault.save(str(path_with_pass), password=password)

    # Files should be different
    raw_no_pass = path_no_pass.read_bytes()
    raw_with_pass = path_with_pass.read_bytes()
    assert raw_no_pass != raw_with_pass

    # Both should load correctly with appropriate methods
    loaded_no_pass = Vault.load(str(path_no_pass))
    loaded_with_pass = Vault.load(str(path_with_pass), password=password)

    assert loaded_no_pass.get_entry("Test").secret == "JBSWY3DPEHPK3PXP"
    assert loaded_with_pass.get_entry("Test").secret == "JBSWY3DPEHPK3PXP"



import pathlib
from typing import Any, Dict
from unittest.mock import patch

import pytest
from typer.testing import CliRunner

from desktop_2fa.cli.helpers import get_password_from_cli
from desktop_2fa.cli.main import app


@pytest.fixture
def fake_ctx() -> Any:
    """Create a fake typer context."""

    class FakeContext:
        def __init__(self) -> None:
            self.obj: Dict[str, Any] = {}

    return FakeContext()


def test_get_password_from_flag(fake_ctx: Any) -> None:
    """Test getting password from --password flag."""
    fake_ctx.obj["password"] = "testpass"
    fake_ctx.obj["interactive"] = True
    assert get_password_from_cli(fake_ctx) == "testpass"


def test_get_password_from_file(fake_ctx: Any, tmp_path: pathlib.Path) -> None:
    """Test getting password from --password-file flag."""
    password_file = tmp_path / "password.txt"
    password_file.write_text("filepass")
    fake_ctx.obj["password_file"] = str(password_file)
    fake_ctx.obj["interactive"] = True
    assert get_password_from_cli(fake_ctx) == "filepass"


def test_get_password_both_flags_error(fake_ctx: Any) -> None:
    """Test error when both --password and --password-file are provided."""
    fake_ctx.obj["password"] = "test"
    fake_ctx.obj["password_file"] = "file"
    fake_ctx.obj["interactive"] = True
    from click.exceptions import Exit

    with pytest.raises(Exit):
        get_password_from_cli(fake_ctx)


def test_get_password_file_not_found(fake_ctx: Any) -> None:
    """Test error when password file does not exist."""
    fake_ctx.obj["password_file"] = "/nonexistent/file"
    fake_ctx.obj["interactive"] = True
    from click.exceptions import Exit

    with pytest.raises(Exit):
        get_password_from_cli(fake_ctx)


def test_get_password_non_interactive_no_password(fake_ctx: Any) -> None:
    """Test error in non-interactive mode when no password provided."""
    fake_ctx.obj["interactive"] = False
    from click.exceptions import Exit

    with pytest.raises(Exit):
        get_password_from_cli(fake_ctx)


def test_get_password_interactive_success(fake_ctx: Any) -> None:
    """Test interactive password prompt."""
    fake_ctx.obj["interactive"] = True
    with patch("typer.prompt") as mock_prompt:
        mock_prompt.side_effect = ["mypass"]
        assert get_password_from_cli(fake_ctx) == "mypass"
        assert mock_prompt.call_count == 1


def test_get_password_interactive_mismatch_retry(fake_ctx: Any) -> None:
    """Test interactive password prompt."""
    fake_ctx.obj["interactive"] = True
    with patch("typer.prompt") as mock_prompt:
        mock_prompt.side_effect = ["pass1"]
        assert get_password_from_cli(fake_ctx) == "pass1"
        assert mock_prompt.call_count == 1


# CLI integration tests using CliRunner

runner = CliRunner()


@pytest.fixture
def fake_vault_env(tmp_path: pathlib.Path, monkeypatch: Any) -> pathlib.Path:
    fake_vault = tmp_path / "vault"

    monkeypatch.setattr(
        "desktop_2fa.cli.helpers.get_vault_path",
        lambda: str(fake_vault),
    )

    # Ensure directory is clean
    if fake_vault.parent.exists():
        import shutil

        shutil.rmtree(fake_vault.parent)
    fake_vault.parent.mkdir(parents=True, exist_ok=True)

    return fake_vault


def test_cli_password_flag(fake_vault_env: pathlib.Path) -> None:
    """Test --password flag in CLI."""
    result = runner.invoke(
        app, ["--password", "testpass", "add", "GitHub", "GitHub", "JBSWY3DPEHPK3PXP"]
    )
    assert result.exit_code == 0
    assert "Entry added: GitHub" in result.output


def test_cli_password_file_flag(
    fake_vault_env: pathlib.Path, tmp_path: pathlib.Path
) -> None:
    """Test --password-file flag in CLI."""
    password_file = tmp_path / "pass.txt"
    password_file.write_text("filepass")
    result = runner.invoke(
        app,
        ["--password-file", str(password_file), "add", "GitHub", "GitHub", "JBSWY3DPEHPK3PXP"],
    )
    assert result.exit_code == 0
    assert "Entry added: GitHub" in result.output


def test_cli_both_password_flags_error(
    fake_vault_env: pathlib.Path, tmp_path: pathlib.Path
) -> None:
    """Test error when both --password and --password-file are provided."""
    password_file = tmp_path / "pass.txt"
    password_file.write_text("filepass")
    result = runner.invoke(
        app, ["--password", "test", "--password-file", str(password_file), "list"]
    )
    assert result.exit_code == 1
    assert "Error: Cannot specify both --password and --password-file" in result.output


def test_cli_non_interactive_no_password_error(fake_vault_env: pathlib.Path) -> None:
    """Test error in non-interactive mode without password."""
    with (
        patch("sys.stdin.isatty", return_value=False),
        patch("sys.stdout.isatty", return_value=False),
    ):
        result = runner.invoke(app, ["list"])
        assert result.exit_code == 1
        assert (
            "Error: Password not provided and not running in interactive mode"
            in result.output
        )


def test_cli_interactive_password_prompt(fake_vault_env: pathlib.Path) -> None:
    """Test interactive password prompt in CLI."""
    # Set environment variable to force interactive mode for testing
    import os

    original_env = os.environ.get("DESKTOP_2FA_FORCE_INTERACTIVE")
    os.environ["DESKTOP_2FA_FORCE_INTERACTIVE"] = "1"

    try:
        with patch("typer.prompt") as mock_prompt:
            mock_prompt.side_effect = ["JBSWY3DPEHPK3PXP", "interpass", "interpass"]
            result = runner.invoke(app, ["add", "GitHub", "GitHub"])
            assert (
                result.exit_code == 0
            ), f"Exit code was {result.exit_code}, output: {result.output}"
            assert "Entry added: GitHub" in result.output
            assert mock_prompt.call_count == 3
    finally:
        # Restore original environment
        if original_env is None:
            os.environ.pop("DESKTOP_2FA_FORCE_INTERACTIVE", None)
        else:
            os.environ["DESKTOP_2FA_FORCE_INTERACTIVE"] = original_env
